# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import os
import shutil
from bzrlib import (
    errors,
    urlutils,
)
from .utils import (
    CacheDisk
)
from bzrlib.branch import Branch
from bzrlib.bzrdir import BzrDir
from bzrlib.plugin import load_plugins
from bzrlib.workingtree import WorkingTree

load_plugins()


class NotABranchError(Exception):
    pass


class NoSuchFileError(Exception):
    pass


class DivergedBranchError(Exception):
    pass


class PermissionDeniedError(Exception):
    pass


class NoWorkingTreeError(Exception):
    pass


class AlreadyControlDirError(Exception):
    pass


class BzrBranch(object):
    "Simplified bazaar branch API"
    def __init__(self, location):
        if location.startswith("file://"):
            location = location[7:]
        if urlutils.is_url(location):
            self.is_local = False
        else:
            if not os.path.isabs(location):
                location = os.path.normpath(os.path.join(os.getcwd(),
                                            location))
            self.is_local = True
        self.location = location
        self._branch = None
        self._tree = None

    @property
    def bzr_branch(self):
        if not self._branch:
            try:
                self._branch = Branch.open(self.location)
            except errors.NotBranchError:
                return None
        return self._branch

    @property
    def bzr_tree(self):
        if self._tree:
            return self._tree
        if self.is_local:
            try:
                self._tree = WorkingTree.open(self.location)
                return self._tree
            except errors.NoWorkingTree:
                return None
        else:
            return None

    @property
    def is_remote(self):
        return not self.is_local

    @property
    def is_branch(self):
        if self.is_local:
            return self.is_local_branch()
        else:
            return self.is_remote_branch()

    def is_local_branch(self):
        if self.bzr_branch is None:
            return False
        return True

    # only cache remote is_branch() calls
    @CacheDisk.cache(unique_attrs=('location',))
    def is_remote_branch(self):
        if self.bzr_branch is None:
            return False
        return True

    @property
    def has_tree(self):
        if self.bzr_tree is None:
            return False
        return True

    @property
    @CacheDisk.cache(unique_attrs=('location',))
    def revno(self):
        if self.bzr_branch:
            return self.bzr_branch.revno()

    @property
    def revno_tree(self):
        if self.is_remote or not self.bzr_tree:
            return None
        return self.bzr_branch.revision_id_to_revno(
            self.bzr_tree.last_revision())

    @property
    def parent(self):
        if self.is_remote or self.bzr_branch is None:
            return None
        return self.bzr_branch.get_parent()

    def has_changes(self, unknowns=False, detritus=False):
        "Check whether changes or stray files are present in the branch"
        if self.is_remote:
            return False
        if not self.bzr_tree:
            return False
        if self.bzr_tree.has_changes():
            return True
        if unknowns and list(self.bzr_tree.unknowns()):
            return True
        if detritus and list(self.bzr_tree.ignored_files()):
            return True
        return False

    def revert(self):
        if self.is_remote or not self.bzr_tree:
            raise NoWorkingTreeError
        self.bzr_tree.revert()

    def cleantree(self, unknowns=True, detritus=True):
        if self.is_remote or not self.bzr_tree:
            raise NoWorkingTreeError
        targets = []
        if unknowns:
            targets.extend(list(self.bzr_tree.unknowns()))
        if detritus:
            self.bzr_tree.lock_read()
            targets.extend([f[0] for f in self.bzr_tree.ignored_files()])
        for target in [os.path.join(self.location, t) for t in targets]:
            if os.path.isdir(target):
                shutil.rmtree(target)
            else:
                os.unlink(target)

    def clone(self, dest_dir, revno=None):
        if not self.bzr_branch:
            return None
        if revno is not None:
            revision_id = self.bzr_branch.get_rev_id(revno)
        else:
            revision_id = None
        try:
            self.bzr_branch.bzrdir.sprout(dest_dir, revision_id=revision_id)
        except errors.PermissionDenied:
            raise PermissionDeniedError("Could not create bzr repo in {}"
                                        "".format(dest_dir))
        except errors.NoSuchFile:
            raise NoSuchFileError("No such file or directory: {}"
                                  "".format(dest_dir))
        cloned = BzrBranch(dest_dir)
        CacheDisk.delete_cache_dir(cloned, unique_attrs=('location',))
        return cloned

    @CacheDisk.delete(unique_attrs=('location',))
    def create(self):
        "bzr init"
        if self.is_local:
            try:
                BzrDir.create_branch_convenience(self.location)
            except errors.AlreadyControlDirError:
                raise AlreadyControlDirError

    def pull(self, src_location=None, overwrite=False, revno=None):
        if not self.is_branch:
            raise NotABranchError("{} is not a branch".format(self.location))
        if src_location is None and self.is_local:
            src_location = self.parent
        other = BzrBranch(src_location)
        if not other.is_branch:
            raise NotABranchError("{} is not a branch".format(other.location))
        try:
            if self.bzr_tree:
                self.bzr_tree.pull(other.bzr_branch,
                                   overwrite=overwrite,
                                   stop_revision=revno)
                self.bzr_tree.update(revision=revno)
            else:
                self.bzr_branch.pull(other.bzr_branch,
                                     overwrite=overwrite,
                                     stop_revision=revno)
        except errors.DivergedBranches:
            message = "Branches have diverged: {} and {}".format(
                self.location, other.location)
            raise DivergedBranchError(message)
        return True

    def push(self, dest_location, update=True, update_revno=None,
             overwrite=False):
        dest_branch = BzrBranch(dest_location)
        if not dest_branch.is_branch:
            return None
        try:
            self.bzr_branch.push(dest_branch.bzr_branch, overwrite=overwrite)
        except errors.DivergedBranches:
            raise DivergedBranchError
        if update:
            dest_branch.update(update_revno)
        return dest_branch

    def update(self, revno=None):
        rev_id = None
        if revno is not None:
            rev_id = self.bzr_branch.get_rev_id(revno)
        self.bzr_tree.update(revision=rev_id)
