# Copyright 2014, 2015 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
from __future__ import print_function
import argparse
import logging
import subprocess
import sys
import os
import pkg_resources

import mojo

from .contain import (
    LXCContainer,
    NoContainer,
    NoSuchContainerError
)
from .exceptions import ArgumentError
import mojo.exceptions
import mojo.manifest
import mojo.project
import mojo.spec
import mojo.workspace


def run_phase(args):
    project, workspace = get_project_and_workspace(args)
    os.environ["MOJO_PROJECT"] = project.name
    os.environ["MOJO_ROOT"] = project.mojo_root
    os.environ["MOJO_WORKSPACE"] = workspace.name
    phase = mojo.phase.build(args.subparser_name, **args.options)
    phase.run(project, workspace, args.series, args.stage)


def show_deploy(args):
    "Show rendered juju-deployer configuration"
    project, workspace = get_project_and_workspace(args)
    phase = mojo.phase.build('deploy', **args.options)
    phase.show(project, workspace, args.series, args.stage)


def diff_deploy(args):
    """
    Show juju-deployer diff. Generate a delta between a
    configured deployment and a running environment
    """
    project, workspace = get_project_and_workspace(args)
    phase = mojo.phase.build('deploy', **args.options)
    phase.diff(project, workspace, args.series, args.stage)


class OptionParser(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        try:
            options = dict(option.split("=", 1) for option in values)
        except ValueError:
            print("Options must be key=value pairs.", file=sys.stderr)
            parser.print_help(file=sys.stderr)
            sys.exit(1)
        setattr(namespace, self.dest, options)


# -----------------------
# helpers
# -----------------------
def get_project_and_workspace(args):
    project = mojo.project.Project(get_project_name(args), args.series, args.mojo_root)
    workspace = project.workspace(args.workspace, args.logfile)
    return project, workspace


def get_project_name(args, use_spec=False):
    """Return the project name from args or the environment.

    Priority for setting project: command line switch then
    MOJO_PROJECT environment variable

    If use_spec is true and the above priority settings are
    missing, then the project will be guessed from spec_url.
    A warning will be printed to this effect.

    Raises ArgumentError if no project can be determined.
    """

    # command line switch, --project
    project_name = args.project
    # MOJO_PROJECT environment variable
    if not project_name:
        project_name = os.environ.get("MOJO_PROJECT")
    # Guess based on spec_url and warn
    if not project_name and use_spec:
        spec_url, workspace = get_spec_url_and_workspace(args)
        project_name = spec_url.rsplit("/")[-1]
        if not project_name:
            project_name = args.spec_url.rsplit("/")[-2]
        logging.warn("Guessing the project is {}. Use --project or set "
                     "MOJO_PROJECT to be more explicit.".format(project_name))
    # Fail gracefully
    if not project_name:
        raise ArgumentError("Project setting missing. Use --project or set "
                            "MOJO_PROJECT")
    return project_name


def get_spec_url_and_workspace(args):
    """Return the spec_url and workspace from args or the
    environment.

    Returns spec_url and workspace.

    Raises ArgumentError if either spec_url or workspace
    cannot be determined.
    """
    spec_url = args.spec_url or os.environ.get("MOJO_SPEC")
    workspace = args.workspace or os.environ.get("MOJO_WORKSPACE")

    if not spec_url or not workspace or spec_url == workspace:
        raise ArgumentError("Missing spec_url and/or workspace. Set "
                            "MOJO_SPEC and/or MOJO_WORKSPACE")

    return (spec_url, workspace)


def get_container(args):
    """Return the container from args or the environment.
    Returns 'LXCContainer' by default.
    Currently this is only required at project creation time.
    Stateful project information will be gathered on subsequent
    mojo commands. If this behavior changes, the defaults below
    may need to change.
    """
    if args.container == 'lxc':
        return LXCContainer
    elif args.container == 'containerless':
        return NoContainer
    else:
        raise NoSuchContainerError("{} is an unsupported container type."
                                   "".format(args.container))


# -----------------------
# commands
# -----------------------
def create_project(args):
    "Set up a Mojo project and LXC container"
    project = mojo.project.Project(args.name, args.series, args.mojo_root,
                           container_class=get_container(args))
    project.create()


def destroy_project(args):
    try:
        if args.series:
            series_list = (args.series,)
        else:
            series_list = mojo.project.Project.list_series(args.mojo_root, args.name)

        projects = [mojo.project.Project(args.name, series, args.mojo_root) for series in series_list]

        for project in projects:
            logging.info("Destroying project: {}, series: {}"
                         "".format(project.name, project.series))
            project.delete()

    except mojo.project.ProjectNameError:
        logging.error('Invalid project name: {}'.format(args.name))
        sys.exit(1)
    except mojo.project.ProjectDoesNotExistError:
        logging.error('Invalid project: {}'.format(args.name))
        sys.exit(1)
    except mojo.project.ProjectCreationError:
        logging.exception('Invalid project: {}'.format(args.name))
        sys.exit(1)


def list_projects(args):
    map(print, mojo.project.Project.list(args.mojo_root, only_names=args.names))


def create_workspace(args):
    "Set up workspace and fetch a copy of the spec"
    project = mojo.project.Project(get_project_name(args), args.series, args.mojo_root)
    spec_url, workspace = get_spec_url_and_workspace(args)

    try:
        ws = project.create_workspace(spec_url, workspace,
                                      stage=args.stage,
                                      logfile=args.logfile)
    except OSError as e:
        print("Are you sure the project exists?", file=sys.stderr)
        if args.blunt:
            raise
        sys.exit(2)
    except AttributeError as e:
        print("Unable to branch {}".format(args.spec_url))
        print("Please check the branch exists and you have permissions to "
              "access it")
        # Now we need to clean up the workspace
        project.delete_workspace(workspace)
        sys.exit(2)
    logging.info("Workspace created. Validating manifest")
    try:
        manifest = ws.spec.manifest
        logging.info("Manifest OK ({})".format(manifest.config))
    except mojo.exceptions.ConfigNotFoundException as e:
        logging.error(e.message)
        sys.exit(1)
    except ValueError as e:
        logging.error(e.message)
        sys.exit(1)


def destroy_workspace(args):
    project_name = get_project_name(args)

    try:
        project = mojo.project.Project(project_name, args.series, args.mojo_root)
        project.workspace(args.workspace)
    except mojo.workspace.NoSuchWorkspaceError:
        logging.error(
            'No such workspace (project: {project}; series: {series}): {workspace}'.format(
                project=project_name,
                series=args.series,
                workspace=args.workspace,
            ))
        sys.exit(1)
    except mojo.project.ProjectNameError:
        logging.error('Invalid project name: {}'.format(project_name))
        sys.exit(1)
    except mojo.project.ProjectDoesNotExistError:
        logging.error('Invalid project: {}'.format(project_name))
        sys.exit(1)
    except mojo.project.ProjectCreationError:
        logging.exception('Invalid project: {}'.format(project_name))
        sys.exit(1)

    project.delete_workspace(args.workspace)


def list_workspaces(args):
    project = mojo.project.Project(get_project_name(args), args.series, args.mojo_root)
    map(print, project.workspace_list())


def display_help(args):
    "Display help for sub-commands"
    if not args.command:
        args.parser.print_help()
    else:
        args.parser.parse_args([args.command, '-h'])
    sys.exit(0)


def run_from_manifest(args):
    project_name = get_project_name(args, use_spec=True)
    spec_url, workspace = get_spec_url_and_workspace(args)
    if args.keep_local_spec_changes:
        clean_spec = False
    else:
        clean_spec = True
    project = mojo.project.Project(project_name, args.series, args.mojo_root)
    workspace = project.create_workspace(spec_url, workspace,
                                         stage=args.stage,
                                         manifest_file=args.manifest_file,
                                         clean_spec=clean_spec,
                                         logfile=args.logfile)
    manifest = workspace.spec.manifest
    manifest.run(project, workspace, args.series, args.stage)


def create_arg_parser():
    # main program parser
    ap = argparse.ArgumentParser()
    ap.add_argument("--blunt", action="store_true",
                    help="Handle fewer exceptions")
    ap.add_argument("--log_level", "-l", default='INFO',
                    help="Log level. Defaults to INFO")
    ap.add_argument("--logfile", "-L", default=os.environ.get('MOJO_LOGFILE'),
                    help="File to log to in addition to stdout. Overrides environment "
                         "variable MOJO_LOGFILE. Defaults to MOJO_WORKSPACE/log/mojo.log")
    ap.add_argument("-r", "--mojo_root", default=os.environ.get('MOJO_ROOT', '/srv/mojo'),
                    help="Root directory that mojo works out of Overrides the MOJO_ROOT environment variable."
                         "Defaults to /srv/mojo")
    ap.add_argument('--version', action='version',
                    version=pkg_resources.require("mojo")[0].version)

    # subcommand parser
    sp = ap.add_subparsers(title="subcommands", dest="subparser_name")

    # Common subcommand args, set here in one place and then reuse in subcommands as needed
    project_args = ("-p", "--project")
    project_kwargs = {'help': "Project name. Overrides the MOJO_PROJECT environment variable."}
    series_args = ("-s", "--series")
    series_kwargs = {'default': os.environ.get('MOJO_SERIES'),
                     'help': "Distro series. Overrides the MOJO_SERIES environment variable."}
    stage_args = ("--stage", )
    stage_kwargs = {'default': os.environ.get('MOJO_STAGE', 'devel'),
                    'help': "Deployment stage. Overrides MOJO_STAGE environment variable (default: devel)"}
    workspace_args = ("-w", "--workspace")
    workspace_kwargs = {'default': os.environ.get('MOJO_WORKSPACE'),
                        'help': "Workspace name. Overrides the MOJO_WORKSPACE environment variable."}

    # options parser for phase commands
    phaseopts = argparse.ArgumentParser(add_help=False)
    phaseopts.add_argument(*stage_args, **stage_kwargs)
    phaseopts.add_argument("--options",
                           help="Phase options. i.e. config=CONFIG_FILENAME",
                           nargs="*", metavar="key=value", action=OptionParser,
                           default={})
    phaseopts.add_argument(*project_args, **project_kwargs)
    phaseopts.add_argument(*workspace_args, **workspace_kwargs)
    phaseopts.add_argument(*series_args, **series_kwargs)

    # project-new subcommand - creates a new mojo project
    prjnewcmd = sp.add_parser("project-new",
                              description="Create a new Mojo project and container")
    prjnewcmd.add_argument(*series_args, required=False, **series_kwargs)
    prjnewcmd.add_argument("name",
                           help="Base name of the project ([a-zA-Z0-9._]*)")
    prjnewcmd.add_argument('-c', "--container", default=os.environ.get('MOJO_CONTAINER', 'lxc'),
                           choices=['lxc', 'containerless'],
                           help="Container. Options 'lxc' or 'containerless'. "
                                "Defaults to 'lxc'")
    prjnewcmd.set_defaults(func=create_project)

    # project-destroy subcommand - deletes a mojo project
    prjdelcmd = sp.add_parser("project-destroy",
                              description="Destroy a Mojo project and container")
    prjdelcmd.add_argument("name",
                           help="Base name of the project ([a-zA-Z0-9._]*)")
    prjdelcmd.add_argument(*series_args, **series_kwargs)
    prjdelcmd.set_defaults(func=destroy_project)

    # project-list
    prjlistcmd = sp.add_parser("project-list",
                               description="List existing projects, found at the mojo root")
    prjlistcmd.add_argument("-n", "--names", action="store_true",
                            help="List only project names, without series")
    prjlistcmd.set_defaults(func=list_projects)

    # workspace-new subcommand - creates a new workspace for a specific spec
    # revision
    wsnewcmd = sp.add_parser("workspace-new",
                             description="Initialize a mojo workspace")
    wsnewcmd.add_argument("spec_url", nargs='?', default=None,
                          help="Bzr URL to the Mojo spec. Overrides the "
                               "MOJO_SPEC environment variable.")
    wsnewcmd.add_argument("workspace", nargs='?', default=None, help=workspace_kwargs['help'])
    wsnewcmd.add_argument(*project_args, **project_kwargs)
    wsnewcmd.add_argument(*series_args, **series_kwargs)
    wsnewcmd.add_argument(*stage_args, **stage_kwargs)
    wsnewcmd.set_defaults(func=create_workspace, parser=wsnewcmd)

    # workspace-destroy subcommand - deletes a mojo project
    wsdelcmd = sp.add_parser("workspace-destroy",
                             description="Destroy a mojo workspace")
    wsdelcmd.add_argument(*project_args, **project_kwargs)
    wsdelcmd.add_argument(*series_args, **series_kwargs)
    wsdelcmd.add_argument(*workspace_args, **workspace_kwargs)
    wsdelcmd.set_defaults(func=destroy_workspace)

    # workspace-list subcommand
    wslistcmd = sp.add_parser("workspace-list",
                              description="List workspaces for a project")
    wslistcmd.add_argument(*project_args, **project_kwargs)
    wslistcmd.add_argument(*series_args, **series_kwargs)
    wslistcmd.set_defaults(func=list_workspaces)

    # builddeps subcommand - install build dependencies for project
    depscmd = sp.add_parser("builddeps", description="builddeps",
                            parents=[phaseopts])
    depscmd.set_defaults(func=run_phase)

    # collect subcommand - collects code branches to build directory
    clctcmd = sp.add_parser("collect",
                            description="Collect resources for build",
                            parents=[phaseopts])
    clctcmd.set_defaults(func=run_phase)

    # build subcommand - run project build script inside container
    buildcmd = sp.add_parser("build",
                             description="Execute project build script",
                             parents=[phaseopts])
    buildcmd.set_defaults(func=run_phase)

    # repo subcommand - assembles charms into a repository
    repocmd = sp.add_parser("repo", description="Assemble a charm repository",
                            parents=[phaseopts])
    repocmd.set_defaults(func=run_phase)

    # sleep subcommand - sleeps for n seconds
    sleepcmd = sp.add_parser("sleep", description="Sleep for time in seconds",
                             parents=[phaseopts])
    sleepcmd.set_defaults(func=run_phase)

    # secrets subcommand - secretss for n seconds
    secretscmd = sp.add_parser("secrets",
                               description="Copy secrets from staged "
                               "location to workspace local",
                               parents=[phaseopts])
    secretscmd.set_defaults(func=run_phase)

    # script subcommand - run an arbitrary script inside container
    scriptcmd = sp.add_parser("script", description="Execute a project script",
                              parents=[phaseopts])
    scriptcmd.set_defaults(func=run_phase)

    # deploy subcommand - run a juju-deployer session
    deploycmd = sp.add_parser("deploy",
                              description="Deploy services and/or create "
                              "relations",
                              parents=[phaseopts])
    deploycmd.add_argument("--ignore", action="store_true", default=False,
                           help="Exit 0 even if service deployment fails")
    deploycmd.set_defaults(func=run_phase)

    # deploy-show subcommand
    deployshowcmd = sp.add_parser("deploy-show",
                                  description="Show rendered juju-deployer "
                                              "configuration",
                                  parents=[phaseopts])
    deployshowcmd.set_defaults(func=show_deploy)

    # deploy-diff subcommand
    deploydiffcmd = sp.add_parser("deploy-diff",
                                  description="Show juju-deployer diff. "
                                              "Generate a delta between a "
                                              "configured deployment and a "
                                              "running environment.",
                                  parents=[phaseopts])
    deploydiffcmd.set_defaults(func=diff_deploy)

    # volumes subcommand - mount volumes
    volcmd = sp.add_parser("volumes", description="Mount volumes",
                           parents=[phaseopts])
    volcmd.set_defaults(func=run_phase)

    # verify subcommand - verify that a deployment was successful
    verifycmd = sp.add_parser("verify",
                              description="Verify that the deployment is "
                              "working",
                              parents=[phaseopts])
    verifycmd.set_defaults(func=run_phase)

    # juju-check-wait subcommand - wait until juju environment reaches steady state
    jujucheckwaitcmd = sp.add_parser("juju-check-wait",
                                description="Wait until juju environment "
                                            "reaches steady state.",
                                parents=[phaseopts])
    jujucheckwaitcmd.set_defaults(func=run_phase)

    # run subcommand - run an entire spec per its manifest
    runcmd = sp.add_parser("run",
                           description="Run an entire deployment from start "
                           "to finish")
    runcmd.add_argument("spec_url", nargs='?', default=None,
                        help="Bzr URL to the Mojo spec. Overrides the "
                             "MOJO_SPEC environment variable.")
    runcmd.add_argument("workspace", nargs='?', default=None, help=workspace_kwargs['help'])
    runcmd.add_argument("--manifest-file", default="manifest",
                        help="The particular manifest file to run")
    runcmd.add_argument(*series_args, **series_kwargs)
    runcmd.add_argument(*stage_args, **stage_kwargs)
    runcmd.add_argument(*project_args, **project_kwargs)
    runcmd.add_argument("-k", "--keep-local-spec-changes", help="Keep "
                        "local edits. This option allows you to make local "
                        "edits in $MOJO_SPEC_DIR for quick testing",
                        action="store_true")
    runcmd.set_defaults(func=run_from_manifest)

    # help subcommand - get help for other subcommands
    helpcmd = sp.add_parser("help", description="Interactive help")
    helpcmd.add_argument("command", nargs='?', default=None,
                         help="Get help for this sub-command.")
    helpcmd.set_defaults(func=display_help, parser=ap)

    return ap


def run_with_args(args, ap=None):
    if hasattr(args, "cache_ttl") and args.cache_ttl is not None:
        mojo.CacheDisk.ttl(args.cache_ttl)
    try:
        args.func(args)
    except ArgumentError as e:
        logging.error(e.message)
        if ap is not None:
            ap.parse_args([args.subparser_name, '-h'])
        sys.exit(1)
    except mojo.utils.Unprivileged:
        logging.error("This sub-command, {}, must be run as root"
                      "".format(args.subparser_name))
        sys.exit(1)
    except (mojo.juju.status.JujuStatusError,
            mojo.jujuwait.JujuWaitException,
            mojo.phase.CollectPhaseException,
            mojo.phase.StopOnPhaseException,
            mojo.phase.VerifyPhaseException,
            mojo.phase.JujuDeployerException) as e:
        logging.error(e)
        sys.exit(1)
    except mojo.phase.ScriptPhaseException as e:
        logging.error(e.output)
        sys.exit(1)
    except mojo.exceptions.ConfigNotFoundException as e:
        logging.error(e.message)
        print("Searched:", file=sys.stderr)
        for path in e.search_list:
            print("  {}".format(path), file=sys.stderr)
        sys.exit(1)
    except (mojo.exceptions.InvalidManifestException,
            mojo.charm_repo.DuplicateCharmError,
            mojo.phase.FileDoesNotExistException,
            mojo.phase.InvalidStopOnReturnCodeException,
            mojo.phase.InvalidYAMLException) as e:
        logging.error(e.message)
        sys.exit(1)
    except subprocess.CalledProcessError:
        logging.exception("Error calling external process")
        sys.exit(1)
    except Exception as e:
        logging.exception("Unknown error")
        sys.exit(1)


UNWANTED_ENVIRONMENT_VARIABLES = [
    # We expect juju-deployer to look in the cwd,
    # but setting $JUJU_REPOSITORY overrides that.
    'JUJU_REPOSITORY',
]


def sanitize_environment():
    for variable in UNWANTED_ENVIRONMENT_VARIABLES:
        if variable in os.environ:
            logging.warning(
                'Removing ${} from the environment.'.format(variable)
            )
            del(os.environ[variable])


# -----------------------
# main
# -----------------------


def main():
    ap = create_arg_parser()
    args = ap.parse_args()
    logFormatter = logging.Formatter(
        fmt="%(asctime)s [%(levelname)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S")
    rootLogger = logging.getLogger()
    rootLogger.setLevel(args.log_level)

    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(logFormatter)
    rootLogger.addHandler(consoleHandler)

    sanitize_environment()

    run_with_args(args, ap)

if __name__ == "__main__":
    sys.exit(main())
