# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
#
# check for LXC container for workspace
# create LXC container for workspace
# execute scripts inside of container
# enable or disable networking inside the container

import os
import subprocess
import sys
from contextlib import contextmanager
import shlex
import shutil
import tempfile
import logging
from .utils import needs_root
import yaml
import pwd
import platform
import apt_pkg

UBUNTU_RELEASES = {'precise': '12.04',
                   'trusty': '14.04',
                   'utopic': '14.10',
                   'vivid': '15.04',
                   'wily': '15.10'
                   }

ARCHITECTURES = {'32bit': 'i386',
                 '64bit': 'amd64'
                 }


class NoSuchContainerError(Exception):
    pass


class InvalidConfigurationError(Exception):
    pass


class InvalidUbuntuSeries(Exception):
    pass


# Decorator for methods that require a defined container
def ensure_defined(method):
    def wrapped(self, *args, **kwargs):
        if not self.defined:
            raise NoSuchContainerError(self.container_name)
        return method(self, *args, **kwargs)
    return wrapped


class Container(object):
    def __init__(self, name, series, ensure_exists=False):
        self.name = name
        self.series = series
        self.container_name = "{}.{}".format(name, series)
        self.lxc_version_validated = None
        if type(self) == LXCContainer:
            self.container_class = 'lxc'
            self.container_root = "/var/lib/lxc/{}/rootfs" \
                                  "".format(self.container_name)
        elif type(self) == NoContainer:
            self.container_class = 'containerless'
            self.container_root = ''
        if ensure_exists and not self.defined:
            raise NoSuchContainerError(self.container_name)
        if self.defined:
            self.load_config()
        self.utils = ContainerUtils(self)
        if not UBUNTU_RELEASES.get(self.series):
            raise InvalidUbuntuSeries("{} is not a valid Ubuntu series. It "
                                      "may no longer be supported or it may "
                                      "be new and needs to be added to mojo."
                                      "".format(self.series))

    def load_config(self):
        pass

    @contextmanager
    def running_container(self, network=True, keep_running=False):
        running = self.running
        if not running:
            self.start(network=network)
        try:
            yield
        finally:
            if not running and not keep_running:
                self.stop()

    def create(self):
        raise NotImplemented

    def destroy(self, force=False):
        raise NotImplemented

    def start(self):
        raise NotImplemented

    def stop(self):
        raise NotImplemented

    def run(self, command, env=None, user=None, network=True, stdout=None,
            stderr=None):
        raise NotImplemented

    @property
    def defined(self):
        return False

    @property
    def running(self):
        return False

    @property
    def container_path(self):
        return None

    @property
    def rootfs(self):
        return None

    def lxc_version(self):
        apt_pkg.init()
        pkgcache = apt_pkg.Cache()
        pkg = pkgcache['lxc']
        return pkg.current_ver.ver_str

    def validate_lxc_version(self, min_version='1.0.0'):
        """LXC >= 1.0 supports features like unprivileged conainers and
        --dir. Return True if the installed version of LXC is >= min_version
        """
        if self.lxc_version_validated is None:
            result = apt_pkg.version_compare(self.lxc_version(), min_version)
            if result >= 0:
                self.lxc_version_validated = True
            else:
                self.lxc_version_validated = False
        return self.lxc_version_validated


class LXCContainer(Container):

    @ensure_defined
    def load_config(self):
        self.config = {}
        is_config_line = lambda l: l.strip() and not l.strip().startswith("#")
        with open(self.conf) as cf:
            # ignore comments
            for line in cf:
                if is_config_line(line):
                    k, v = line.split("=", 1)
                    self.config[k.strip()] = v.strip()

    @needs_root
    def create(self):
        if self.defined:
            raise InvalidConfigurationError("Duplicate container: {}"
                                            "".format(self.container_name))
        # we want to use the ubuntu template so that the environment is
        # consistent with what users expect, but we also make our own
        # adjustments to it as part of Project setup. It might be worth
        # building our own LXC template for Mojo at some point.
        dist = platform.dist()[0].lower()
        arch = ARCHITECTURES.get(platform.architecture()[0])
        if UBUNTU_RELEASES.get(self.series) >= UBUNTU_RELEASES['vivid']:
            template = 'download'
        else:
            template = dist
        cmd = ("lxc-create",
               "-t", template,
               "-n", self.container_name,
               )
        # Can't use lxc location on precise's version of lxc missing -B --dir
        # option
        if os.path.isdir(os.path.dirname(self.container_root)) \
                and self.validate_lxc_version():
            cmd += ("-B", "dir", "--dir", self.container_root)
        cmd += ("--",)
        if self.series is not None:
            cmd += ("--release", self.series)
        if UBUNTU_RELEASES.get(self.series) >= UBUNTU_RELEASES['vivid']:
            cmd += ("--dist", dist)
            cmd += ("--arch", arch)
        logging.info("Creating LXC container '{}'".format(self.container_name))
        logging.info("Running: '{}'".format(cmd))
        subprocess.check_call(cmd)
        logging.info("LXC creation Complete")

    @needs_root
    def setup(self):
        pkgs_to_install = ['software-properties-common',
                           'python-software-properties',
                           'rsync']
        # lxc < 1.0 and the precise kernel do not have a feature needed to use
        # lxc-start, lxc-execute requires that lxc be installed inside the lxc
        # container.
        if not self.validate_lxc_version():
            pkgs_to_install.append('lxc')
        logging.info("Bootstrapping lxc env")
        self.load_config()
        # Serge Hallyn's solution to bootstrapping uses custom init scripts
        # The first init script runs whenever the lxc is started and it updates
        # and installs lxc and other packages needed
        with open(os.path.join(self.rootfs, 'etc', 'init',
                  'containersetup.conf'),
                  'w') as containersetup:
            containersetup.write(
                '''description "container setup job"
author "Serge Hallyn <serge.hallyn@canonical.com>"

start on runlevel [2345]
script
   apt-get update
   apt-get -y dist-upgrade
   apt-get -y install {}
end script\n'''.format(" ".join(pkgs_to_install)))
        # The second init script runs after the first has completed and
        # stops the lxc signaling completion of the installs
        # It also stops itself from running again by echoing 'manual'
        # into itself
        with open(os.path.join(self.rootfs, 'etc', 'init',
                  'setupstopped.conf'), 'w') as setupstopped:
            setupstopped.write(
                '''description "stop lxc containersetup has stopped"
author "Serge Hallyn <serge.hallyn@canonical.com>"
start on stopped containersetup

script
echo manual >> /etc/init/setup-stopped.conf
/sbin/poweroff
end script\n''')
        # Starting the lxc will run the init scripts
        self.start()
        # Wait for the lxc to stop signaling the completion of the installs
        self.wait("STOPPED")
        logging.info("Finished Bootstrapping")

        logging.info("Creating network-disabled and nested lxc configurations")
        shutil.copy(self.conf, self.conf_virtnet)
        with open(self.conf_virtnet) as virtnet:
            with open(self.conf_nonet, "w") as nonet, open(self.conf,
                                                           "w") as hostnet:
                for line in virtnet:
                    # Handle cgroups in a nested lxc per Serge Hallyn
                    line = line.replace(
                        "#lxc.aa_profile = lxc-container-default-with-nesting",
                        "lxc.aa_profile = lxc-container-default-with-nesting")
                    if not line.strip().startswith("lxc.network."):
                        hostnet.write(line)
                        nonet.write(line)
                nonet.write("lxc.network.type=empty\n")
        self.load_config()
        logging.info("LXC Setup Complete")

    def exists(self):
        return os.path.exists(self.conf)

    def is_setup(self):
        if UBUNTU_RELEASES.get(self.series) >= UBUNTU_RELEASES['vivid']:
            return True
        setup_file = self.container_root + "/etc/init/setup-stopped.conf"
        return os.path.exists(setup_file)

    @needs_root
    @ensure_defined
    def destroy(self, force=False):
        cmd = ("lxc-destroy", "-n", self.container_name)
        if force:
            cmd += ("-f",)
        subprocess.check_output(cmd)

    @needs_root
    @ensure_defined
    def add_mount(self, host_path, guest_path=None):
        logging.info("Adding host mount to LXC container '{}'"
                     "".format(self.container_name))
        if not os.path.exists(host_path):
            raise InvalidConfigurationError("Host path {} does not exist."
                                            "".format(host_path))
        if not guest_path:
            dest_path = self.rootfs + host_path
        else:
            dest_path = self.rootfs + guest_path
        os.makedirs(dest_path)
        fstab_line = "{} {} none bind 0 0".format(host_path, dest_path)
        subprocess.check_output("echo {} >> {}".format(fstab_line, self.fstab),
                                shell=True)

    @needs_root
    @ensure_defined
    def remove_mount(self, host_path):
        logging.info("Removing host mount from LXC container '{}'"
                     "".format(self.container_name))
        cmd = "sed --in-place -e '#{}#d' {}".format(host_path, self.fstab)
        subprocess.check_output(cmd, shell=True)

    def check_state(self, state=None):
        if state:
            cmd = ('sudo', 'lxc-info', '-n', self.container_name)
            containers = yaml.safe_load(subprocess.check_output(cmd))
            # lxc-info shows 'state' (precise) or 'State' (newer)
            return state in containers.get('state', containers.get('State', {}))
        cmd = ('sudo', 'lxc-ls', '-1')
        containers = subprocess.check_output(cmd).split()
        return self.container_name in containers

    @ensure_defined
    def start(self, network=True):
        if self.running:
            logging.info("LXC container '{}' is already running"
                         "".format(self.container_name))
        else:
            logging.info("Starting LXC container '{}'".format(
                self.container_name))
            # always start containers in the background
            cmd = ["sudo", "lxc-start", "-d", "-n", self.container_name]
            keys = {}
            if not network:
                cmd += ["-f", self.conf_nonet]
            cmd += ["-s {}={}".format(k, v) for k, v in keys.items()]
            logging.debug("running {}".format(cmd))
            subprocess.check_call(cmd)
            self.wait("RUNNING")

    @ensure_defined
    def stop(self, wait=True):
        if not self.running:
            logging.info("LXC container '{}' is already stopped"
                         "".format(self.container_name))
        else:
            logging.info("Stopping LXC to container '{}'".format(
                self.container_name))
            cmd = ("sudo", "lxc-shutdown")
            if wait:
                cmd += ("-w",)
            cmd += ("-n", self.container_name)
            subprocess.check_call(cmd)

    @ensure_defined
    def wait(self, state='RUNNING'):
        cmd = ("sudo", "lxc-wait", "-n", self.container_name, "-s", state)
        logging.info("Waiting for LXC container to get to state '{}'"
                     "".format(state))
        subprocess.check_call(cmd)

    @ensure_defined
    def run(self, command, env=None, user=None, network=True, stdout=None,
            stderr=None):
        "Run commands in the LXC environment"
        if not command:
            logging.warn("Refusing to execute empty command")
            return -1
        if not user:
            if os.environ.get('SUDO_USER'):
                user = os.environ.get('SUDO_USER')
            else:
                user = pwd.getpwuid(os.getuid()).pw_name
        # Are we attached to a tty? If not we'll need to create a logfile
        # for later use
        tty_attached = sys.stdout.isatty()
        if not tty_attached:
            lxc_console_log = tempfile.NamedTemporaryFile(delete=False)
            lxc_console_log.close()
        # Avoid looping fork bomb by not running when called as user
        # root
        if user != 'root':
            if not self.utils.user_exists(user):
                self.utils.create_user(user)
        command = ["sudo", "-E", "-u", user] + shlex.split(command)
        if isinstance(env, dict):
            env_vars = ["env"] + ["=".join((k, v)) for k, v in env.items()]
            command = env_vars + command
        if self.validate_lxc_version():
            cmd = ["sudo", "lxc-start", "-n", self.container_name, "-F"]
            if not tty_attached:
                cmd.extend(["-L", lxc_console_log.name])
            if network:
                cmd.extend(["--share-net", "1"])
        else:
            cmd = ["sudo", "lxc-execute", "-n", self.container_name]
            if not network:
                cmd.extend(["-f", self.conf_nonet])
        cmd.extend(["--"])
        cmd.extend(command)
        logging.info("Running command in container '{}': {}"
                     "".format(self.container_name, ' '.join(command)))
        try:
            output = subprocess.check_output(cmd, stderr=stderr)
            if output.strip():
                logging.info(output)
        finally:
            # Output not yet displayed if we don't have a tty
            if not tty_attached:
                try:
                    with open(lxc_console_log.name, 'r') as clog:
                        logging.info(clog.read())
                    os.remove(lxc_console_log.name)
                except OSError:
                    logging.error("Unable to read or delete output of commands run inside LXC")
        # subprocess.check_output will raise an error if it has a non-zero
        # return code in any case, so we can just return 0 here.
        return 0

    @property
    def defined(self):
        return self.check_state()

    @property
    def running(self):
        return self.check_state("RUNNING")

    @property
    def container_config_path(self):
        return os.path.join('/var/lib/lxc', self.container_name)

    @property
    @ensure_defined
    def rootfs(self):
        return self.config['lxc.rootfs']

    @property
    def conf(self):
        return os.path.join(self.container_config_path, "config")

    @property
    def conf_virtnet(self):
        return os.path.join(self.container_config_path, "config.virtnet")

    @property
    def conf_nonet(self):
        return os.path.join(self.container_config_path, "config.nonet")

    @property
    def fstab(self):
        return os.path.join(self.container_config_path, "fstab")


class ContainerUtils(object):
    def __init__(self, container):
        super(ContainerUtils, self).__init__()
        self.container = container

    def add_repo_key(self, repo_key):
        command = 'apt-key add {}'.format(repo_key)
        self.container.run(command, user="root")

    def add_repo(self, repo_spec, confirm=True):
        if confirm:
            confirmation = "-y"
        else:
            confirmation = ""
        command = 'add-apt-repository {} "{}"'.format(confirmation, repo_spec)
        self.container.run(command, user="root")
        self.apt_update()

    def apt_update(self):
        with open(os.devnull, 'w') as devnull:
            self.container.run("env DEBIAN_FRONTEND=noninteractive "
                               "apt-get update", user="root", stdout=devnull)
            self.container.run("env DEBIAN_FRONTEND=noninteractive "
                               "apt-get -y dist-upgrade", user="root",
                               stdout=devnull)

    def install_packages(self, *packages, **kwargs):
        prompt = kwargs.get("prompt", False)
        if prompt:
            assume_yes = ""
        else:
            assume_yes = "-y"
        command = "apt-get {} install {}" "".format(assume_yes,
                                                    " ".join(packages))
        self.container.run(command, user="root")

    def user_exists(self, user):
        command = "getent passwd {}".format(user)
        try:
            with open(os.devnull, 'w') as devnull:
                self.container.run(command, user="root", stdout=devnull)
            return True
        except:
            return False

    def create_user(self, user):
        (username, shadow, uid,
         gid, name, home, shell) = subprocess.check_output(
            ['getent', 'passwd', user]).split(':')
        # if the uid exists update the name to match the caller
        try:
            command = "useradd --home {} --create-home --uid {} {}" \
                "".format(home, uid, username)
            self.container.run(command, user="root")
        except subprocess.CalledProcessError as e:
            if e.returncode == 4:
                command = "cp /etc/sudoers.d/ubuntu-all " \
                    "/etc/sudoers.d/{}-all".format(username)
                self.container.run(command, user="root")
                command = 'sed -i "s/ubuntu/{}/g" ' \
                    '/etc/sudoers.d/{}-all'.format(username, username)
                self.container.run(command, user="root")
                command = 'usermod -l {} -G sudo ubuntu'.format(username)
                self.container.run(command, user="root")
            else:
                raise e


class NoContainer(Container):

    def create(self):
        logging.info("Using no container. Not creating a container")

    def destroy(self, force=False):
        logging.info("Using no container. Not destroying a container")

    def run(self, command, env=None, user=None, network=True, stdout=None,
            stderr=None):
        if isinstance(env, dict):
            env_vars = ["env"] + ["=".join((k, v)) for k, v in env.items()]
            command = env_vars + shlex.split(command)
        logging.warn("Atempting to run in a container with a containerless "
                     "project. Running command without a container: {}"
                     "".format(command))
        output = subprocess.check_output(command, stderr=subprocess.STDOUT)
        if output.strip():
            logging.info(output)

    def exists(self):
        return os.path.exists(self.container_root)

    def is_setup(self):
        return self.exists()

    @property
    def defined(self):
        return True

    @property
    def conf(self):
        logging.info("Containerless project has no container config")
        return self.container_root
