import datetime
import logging
import subprocess
import time
import yaml

from collections import Counter


JUJU_STATUSES = {
    'good': ['ACTIVE', 'started'],
    'bad': ['error'],
    'transitional': ['pending', 'down', 'installed', 'stopped'],
}


class JujuStatusError(Exception):
    pass


class Status(object):
    def __init__(self, environment=None):
        self.environment = environment
        self._status_output = None

    def _get_status(self, force_update=False):
        if self._status_output and not force_update:
            return self._status_output
        cmd = ['juju', 'status', '--format=yaml']
        if self.environment:
            cmd.extend(['-e', self.environment])
        self._status_output = subprocess.check_output(cmd)
        return self._status_output

    def yaml_status(self, force_update=False):
        return yaml.safe_load(self._get_status(force_update))

    def status(self, force_update=False):
        return self._get_status(force_update)

    def _get_machine_state(self, state_type):
        states = Counter()
        for machine_no in self.yaml_status()['machines']:
            if state_type in self.yaml_status()['machines'][machine_no]:
                state = self.yaml_status()['machines'][machine_no][state_type]
            else:
                state = 'unknown'
            states[state] += 1
        return states

    def _get_machine_agent_states(self):
        return self._get_machine_state('agent-state')

    def _get_machine_instance_states(self):
        return self._get_machine_state('instance-state')

    def _get_service_agent_states(self):
        service_state = Counter()
        for service in self.yaml_status()['services']:
            if 'units' in self.yaml_status()['services'][service]:
                for unit in self.yaml_status()['services'][service]['units']:
                    unit_info = self.yaml_status()['services'][service]['units'][unit]
                    service_state[unit_info['agent-state']] += 1
                    if 'subordinates' in unit_info:
                        for sub_unit in unit_info['subordinates']:
                            sub_sstate = \
                                unit_info['subordinates'][sub_unit]['agent-state']
                            service_state[sub_sstate] += 1
        return service_state

    def _get_units_in_state(self, state):
        unit_names = []
        for service in self.yaml_status()['services']:
            if 'units' in self.yaml_status()['services'][service]:
                for unit in self.yaml_status()['services'][service]['units']:
                    unit_info = self.yaml_status()['services'][service]['units'][unit]
                    if unit_info['agent-state'] == state:
                        unit_names.append(unit)
                    if 'subordinates' in unit_info:
                        for sub_unit in unit_info['subordinates']:
                            sub_sstate = \
                                unit_info['subordinates'][sub_unit]['agent-state']
                            if sub_sstate == state:
                                unit_names.append(sub_unit)
        return unit_names

    def check_and_wait(self, timeout=1800):
        checks = {
            'Machines': [{
                         'Heading': 'Instance State',
                         'check_func': self._get_machine_instance_states,
                         },
                         {
                         'Heading': 'Agent State',
                         'check_func': self._get_machine_agent_states,
                         }],
            'Services': [{
                         'Heading': 'Agent State',
                         'check_func': self._get_service_agent_states,
                         }]
        }
        check_and_wait_start = datetime.datetime.now()
        stable_state = [False]
        while False in stable_state:
            check_and_wait_running = datetime.datetime.now() - check_and_wait_start
            if check_and_wait_running.total_seconds() > int(timeout):
                logging.error(subprocess.check_output(
                    ['juju', 'status', '--format=tabular']))
                raise JujuStatusError("Timed out checking Juju status for stable state")
            stable_state = []
            for juju_objtype, check_info in checks.iteritems():
                for check in check_info:
                    check_function = check['check_func']
                    states = check_function()
                    if self._error_check(states):
                        logging.error(subprocess.check_output(
                            ['juju', 'status', '--format=tabular']))
                        for unit in self._get_units_in_state("error"):
                            logging.error("Last 200 lines in %s unit log", unit)
                            logging.error(subprocess.check_output(
                                ['juju', 'ssh', unit, 'sudo', 'tail', '-200',
                                 '/var/log/juju/unit-%s.log' %
                                    unit.replace('/', '-')])
                            )
                        raise JujuStatusError("Error in juju status on {}".format(
                            ", ".join(self._get_units_in_state("error"))))
                    stable_state.append(self._all_stable(states))
            time.sleep(5)
            self._get_status(force_update=True)
        for juju_objtype, check_info in checks.iteritems():
            for check in check_info:
                check_function = check['check_func']
                states = check_function()
                self._summary(juju_objtype, check['Heading'], states)

    def _all_stable(self, states):
        for state in states:
            if state in JUJU_STATUSES['transitional']:
                logging.info('Some statuses are in a transitional state')
                return False
        logging.info('Statuses are in a stable state')
        return True

    def _error_check(self, states):
        for state in states:
            if state in JUJU_STATUSES['bad']:
                logging.error('Some statuses are in a bad state')
                return True
        logging.info('No statuses are in a bad state')
        return False

    def _summary(self, heading, statetype, states):
        output = [heading]
        output.append("   " + statetype)
        for state in states:
            output.append("    %s: %i" % (state, states[state]))
        logging.info("\n".join(output))

    def _service_info(self, svc):
        if svc == "ALL":
            units = {}
            for service in self.yaml_status()['services'].values():
                for unit, data in service.get('units', {}).items():
                    units[unit] = data
        elif svc not in self.yaml_status()['services']:
            return None
        else:
            service = self.yaml_status()['services'][svc]
            units = service.get('units', {})
        return units

    def service_ips(self, service):
        units = self._service_info(service)
        if units:
            return [u.get('public-address') for u in units.values()
                    if u.get('public-address')]
        return units

    def service_units(self, service):
        units = self._service_info(service)
        if units:
            return units.keys()
        return units

    def service_machine_numbers(self, service):
        units = self._service_info(service)
        if units:
            return [u.get('machine') for u in units.values()
                    if u.get('machine')]
        return units

    def service_instance_ids(self, service):
        machine_numbers = self.service_machine_numbers(service)
        if machine_numbers is None:
            return None
        elif not machine_numbers:
            return []
        return [self.machine_instance_id(m) for m in machine_numbers]

    def machine(self, machine_num):
        if machine_num not in self.yaml_status()['machines']:
            return None
        return self.yaml_status()['machines'][machine_num]

    def machine_ip(self, machine_num):
        machine = self.machine(machine_num)
        if machine is None:
            return None
        return machine.get('dns-name')

    def machine_instance_id(self, machine_num):
        machine = self.machine(machine_num)
        if machine is None:
            return None
        return machine.get('instance-id')
