# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import os
import logging
import subprocess
import charm_repo
import re
import shutil
import time
import string
import sys
import yaml
from .utils import (
    bicommand,
    chdir,
    get_dep_prog,
)
from .jujuwait import (
    wait,
    JujuWaitException,
)
from jinja2 import Template
from jinja2.exceptions import TemplateSyntaxError
import mojo
import mojo.juju.status
import codetree
from deployer.config import ConfigStack


class NoSuchPhaseError(Exception):
    pass


class FileDoesNotExistException(Exception):
    pass


class NonNumericException(Exception):
    pass


class JujuDeployerException(Exception):
    pass


class CollectPhaseException(Exception):
    pass


class InvalidYAMLException(Exception):
    pass


class InvalidStopOnReturnCodeException(Exception):
    pass


class StopOnPhaseException(Exception):
    pass


class VerifyPhaseException(Exception):
    pass


class ScriptPhaseException(Exception):
    def __init__(self, returncode, output):
        self.returncode = returncode
        self.output = output


class Phase(object):
    "Base class for phases"
    name = "not_implemented"

    def __init__(self, **kwargs):
        super(Phase, self).__init__()
        # check for config= then script= then name of phase
        self.config_base = kwargs.get('config',
                                      kwargs.get('script', self.name))
        self.options = kwargs

    def __str__(self):
        return "<{} {} {} {}>".format(self.__class__.__name__, self.name,
                                      self.config_base, self.options)

    def run(self, project, workspace, series, stage=None):
        raise NotImplementedError


class SleepPhase(Phase):
    """Sleep for time in seconds
       Arguments:
           config      number of seconds to sleep
                       Default: 30
    """
    name = "sleep"

    def run(self, project, workspace, series, stage=None):
        "Sleep for time in seconds"
        logging.debug("### Running phase {} with options {} ###"
                      "".format(self.name, self.options))
        if self.config_base == 'sleep':
            self.config_base = 30
        logging.info("Sleeping for {} seconds".format(self.config_base))
        try:
            self.config_base = int(self.config_base)
            time.sleep(self.config_base)
        except:
            raise NonNumericException("Time for sleep phase must be the "
                                      "numeric number of seconds not '{}'"
                                      "".format(self.config_base))


class CommentPhase(Phase):
    """Include a well-deliniated comment in the info logger.

    When writing a spec, comments in a manifest are useful not only when
    reading the manifest, but also in the output when running the manifest.

    For convenience, any line(s) in a manifest beginning with a single
    hash will be converted to a comment. If you require a comment in
    the manifest which isn't useful during running, use a double-hash.

    This phase cannot be run from the command line.

    Arguments:
        text      The text to be displayed.
                  Default: "comment"
    """
    name = "comment"

    def run(self, *args, **kwargs):
        border = "#" * 77
        comment = "\n".join([border, self.options['text'], border])
        logging.info("Manifest comment:\n\n" + comment + "\n\n")


class SecretsPhase(Phase):
    """Pull Secrets into workspace local
       Arguments:
           config      Directory where secrets are located
                       Default: /srv/mojo/LOCAL/$MOJO_PROJECT/$MOJO_STAGE
    """
    name = "secrets"

    def run(self, project, workspace, series, stage=None):
        "Pull Secrets into workspace local"

        mojo_local_dir = os.path.join(project.mojo_root, 'LOCAL',
                                      project.name, stage or 'devel')
        logging.debug("### Running phase {} with options {} ###"
                      "".format(self.name, self.options))
        logging.info("Pulling secrets from {} to {}"
                     "".format(mojo_local_dir, workspace.local_dir))
        if os.path.isdir(mojo_local_dir):
            src_files = os.listdir(mojo_local_dir)
            for file_name in src_files:
                full_file_name = os.path.join(mojo_local_dir, file_name)
                if os.path.isfile(full_file_name):
                    shutil.copy(full_file_name, workspace.local_dir)
                elif os.path.isdir(full_file_name):
                    dest_dir = os.path.join(workspace.local_dir, file_name)
                    if os.path.exists(dest_dir):
                        shutil.rmtree(dest_dir)
                    shutil.copytree(full_file_name,
                                    os.path.join(workspace.local_dir,
                                                 file_name))
                else:
                    raise FileDoesNotExistException("{} is not a file or "
                                                    "directory"
                                                    "".format(full_file_name))
        elif 'auto_secrets' in self.options:
            logging.warn("Automatic secrets phase ran but secrets "
                         "directory {} does not exist!".format(mojo_local_dir))
        else:
            raise FileDoesNotExistException("Secrets directory {} does not "
                                            "exist and secrets phase was "
                                            "specifically requested."
                                            "".format(mojo_local_dir))


class CollectPhase(Phase):
    """Collect code and resources into workspace build directory.
       Arguments:
           config      base name for config-manager file
    """
    name = "collect"

    def run(self, project, workspace, series, stage=None, auto_repo=True):
        "Collect build resources using config-manager"
        logging.debug("### Running phase {} with options {} ###"
                      "".format(self.name, self.options))
        logging.info("Building resource tree")
        env = os.environ.copy()
        env['MOJO_SERIES'] = series
        env['MOJO_STAGE'] = stage or 'devel'
        expand_config_base = string.Template(self.config_base).substitute(env)
        config = workspace.spec.get_config(expand_config_base, stage)
        with chdir(workspace.build_dir):
            try:
                collect_config = codetree.config.Config(config)
                collect_success = collect_config.build(use_concurrent=True)
                if not collect_success:
                    raise CollectPhaseException("Error during collect phase")
            except (codetree.handlers.NotSameBranch,
                    codetree.handlers.NotABranch,
                    codetree.handlers.BranchDiverged) as e:
                raise CollectPhaseException("Codetree: {}".format(e))

        if auto_repo:
            # Run Repo phase automatically
            repo = CharmRepoPhase(auto_repo=True)
            repo.run(project, workspace, series, stage)


class InstallBuildDepsPhase(Phase):
    """Install dependencies into project container
       Arguments:
           repo_keys   comma-separated list of apt repository gpg key
                       file names
           repos       comma-separated list of apt repositories
           packages    comma-separated list of packages to insall
    """
    name = "builddeps"

    def run(self, project, workspace, series, stage=None):
        "Install packages"
        logging.debug("### Running phase {} with options {}"
                      "".format(self.name, self.options))
        logging.info("Installing apt repos and packages")
        utils = project.utils
        repo_keys = self.options.get('repo_keys')
        if repo_keys:
            for repo_key in repo_keys.split(","):
                repo_key = os.path.join(workspace.spec_dir, stage or 'devel',
                                        repo_key)
                utils.add_repo_key(repo_key)
        repos = self.options.get('repos')
        if repos:
            for repo in repos.split(","):
                utils.add_repo(repo)
        packages = self.options.get('packages')
        if packages:
            utils.apt_update()
            utils.install_packages(*packages.split(","))


class VolumesPhase(Phase):
    """Attach persistent volumes to services.
       Arguments:
           config    base name for volumes config file
    """
    name = "volumes"

    def run(self, project, workspace, series, stage=None):
        "Attach volumes to services"
        logging.debug("### Running phase {} with options {}".format(self.name,
                      self.options))
        logging.info("Mounting volumes")
        mojovol = get_dep_prog("mojo-vol")
        config = workspace.spec.get_config(self.config_base, stage)
        cmd = [mojovol, '--config', config]
        output = subprocess.check_output(cmd)
        logging.info(output)


class ScriptPhase(Phase):
    """Execute a script within the build directory.
       Arguments:
           config      base name for script file
           lxc         Boolean to use lxc or not
           KEY=VALUE   set as environment variables for use by the script
    """
    name = "script"

    def run(self, project, workspace, series, stage=None, lxc=False,
            network=True, auto_secrets=True):
        if auto_secrets:
            # Run secrets phase automatically first
            secrets = mojo.phase.build('secrets', **{'auto_secrets': True})
            secrets.run(project, workspace, series, stage)
        logging.debug("### Running phase {} with options {}".format(self.name,
                      self.options))
        logging.info("Running script {}".format(self.config_base))
        # Override use of lxc from options
        if "lxc" in self.options:
            lxc = str2bool(self.options['lxc'])

        "Run a script portion of a deployment process"
        # XXX: Run all scripts in a chroot or LXC environment
        env = {}
        # Set any options to the environment
        for key in self.options:
            env[key] = self.options[key]
        env['MOJO_PROJECT'] = project.name
        env['MOJO_WORKSPACE'] = workspace.name
        env['MOJO_SERIES'] = series
        env['MOJO_WORKSPACE_DIR'] = workspace.path
        env['MOJO_BUILD_DIR'] = workspace.build_dir
        env['MOJO_REPO_DIR'] = workspace.repo_dir
        env['MOJO_SPEC_DIR'] = workspace.spec_dir
        env['MOJO_LOG_DIR'] = workspace.log_dir
        env['MOJO_LOCAL_DIR'] = workspace.local_dir
        env['MOJO_STAGE'] = stage or 'devel'
        script = workspace.spec.get_config(self.config_base, stage)
        if lxc:
            project.container.run(script, env=env, network=network)
        else:
            env_vars = ["env"] + ["=".join((k, v)) for k, v in env.items()]
            command = env_vars + [script]
            try:
                output = subprocess.check_output(command, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                raise ScriptPhaseException(e.returncode, e.output)
            if output.strip():
                logging.info(output)


class BuildPhase(ScriptPhase):
    """Execute a build script within a build directory. The build environment
       is restricted to have no access to networks.
       Arguments:
           config     Base name for build script file
    """
    name = "build"

    def run(self, project, workspace, series, stage, auto_secrets=True):
        "Run a build script"
        # XXX: Remove network access during build phase
        super(BuildPhase, self).run(project, workspace, series, stage,
                                    lxc=True, network=False,
                                    auto_secrets=auto_secrets)


class CharmRepoPhase(Phase):
    """Build a charm repo from paths within the build directory.
       Arguments:
           config    Base name for the charm repo spec
    """
    name = "repo"

    def run(self, project, workspace, series, stage):
        "Build a charm repository"
        logging.debug("### Running phase {} with options {}"
                      "".format(self.name, self.options))
        logging.info("Build a charm repository")
        repo = charm_repo.CharmRepo(workspace.repo_dir)
        if self.options.get('auto_repo'):
            logging.info("Creating charm repo automatically from all charms "
                         "in the build directory.")
            repo.build_all(workspace.build_dir, series)
        else:
            logging.warn("It is no longer necessary to specify a repo phase "
                         "in the manifest nor have a repo config file in the "
                         "spec. The repo phase will run automatically after "
                         "each collect phase and will copy over all charms "
                         "from the build directory to the charms directory.")
            config = workspace.spec.get_config(self.config_base, stage)
            repo.build_from_config(config, workspace.build_dir, series)


class DeployerPhase(Phase):
    """Run juju-deployer.
       Arguments:
           config         Base name for deployer config
           optional=y/N   When deployment fails, ignore it
           local          Base name in local dir for deployer config (secrets)
           juju           Juju environment
           delay          Deploy Delay in seconds
                          Default: 0 seconds
           status-timeout We check juju status after each deploy phase. How
                          long in seconds before we timeout if instances remain
                          in a pending status.
                          Default: 1800 seconds (30 minutes)
           target         The juju deployer configuration target to use.
           timeout        Overall timeout for the deployment
                          Default: 2700 seconds (45 minutes)
           retry          number of times to retry failed units
                          Default: unset (no retry)
           wait           Wait for the environment to reach steady state
                          Default: False
    """
    name = "deploy"

    def _get_target(self, configs):
        """
        Return juju-deployer configuration target
        """
        deployer = get_dep_prog("juju-deployer")
        if "target" in self.options:
            target = self.options.get("target")
        else:
            # get the (first, and assumed only) deployment name
            # from the config file
            for config in configs:
                if os.path.isfile(config):
                    configfile = config
                    break
            cmd = ['python', deployer, '-c', configfile, "-l"]
            target = subprocess.check_output(cmd).split("\n")[0]
        return target

    def _validate_configs(self, configs):
        """
        Check for valid YAML or JSON configs.
        Raise exception if one t fails to parse.
        """
        for config in configs:
            try:
                with open(config) as yamlfile:
                    yaml.safe_load(yamlfile)
            except yaml.parser.ParserError as e:
                raise InvalidYAMLException("Invalid YAML or JSON in "
                                           "juju-deployer configuration file "
                                           "{}:\n{}".format(config, e))
            except yaml.scanner.ScannerError as e:
                raise InvalidYAMLException("Invalid YAML or JSON in "
                                           "juju-deployer configuration file "
                                           "{}:\n{}".format(config, e))

    def _deployer_configs(self, project, workspace, series, stage,
                          auto_secrets=True, cli=True):
        """
        Return list of juju deployer rendered configuration files
        If cli is True, then return list of juju deployer command line
        switches for config files:
        ('-c', '/srv/mojo/$MOJO_PROJECT/$MOJO_SPEC_DIR/$MOJO_STAGE/deploy',
         '-c', '/srv/mojo/LOCAL/$MOJO_PROJECT/$MOJO_STAGE/secrets')
        """
        if auto_secrets:
            # Run secrets phase automatically first
            secrets = SecretsPhase(auto_secrets=True)
            secrets.run(project, workspace, series, stage)

        # Process the deployment configs as a Jinja2 template
        tmpl_vars = {
            'project': project.name,
            'workspace': workspace.name,
            'series': series,
            'spec_dir': workspace.spec_dir,
            'local_dir': workspace.local_dir,
            'build_dir': workspace.build_dir,
            'stage': stage,
            'stage_name': os.path.basename(stage),
        }
        # Set arbitrary variables from options
        for key in self.options:
            if tmpl_vars.get(key):
                logging.warn("Option '{}' value '{}' overrides default '{}'"
                             "".format(key, self.options.get(key),
                                       tmpl_vars.get(key)))
            tmpl_vars[key] = self.options[key]

        configs = ()
        for config_src in workspace.spec.get_configs(self.config_base, stage):
            # XXX Why render configs to the juju repo dir?
            config_dest = os.path.join(workspace.repo_dir, config_src)
            if not os.path.exists(os.path.dirname(config_dest)):
                os.makedirs(os.path.dirname(config_dest))
            self.render_config(
                os.path.join(workspace.spec_dir, config_src), config_dest,
                tmpl_vars)
            configs += (config_dest,)

        if "local" in self.options:
            local_cfg_src = os.path.join(workspace.local_dir,
                                         self.options["local"])
            if not os.path.exists(local_cfg_src):
                raise FileDoesNotExistException("Local config file does not "
                                                "exist: {}"
                                                "".format(local_cfg_src))
            local_cfg = os.path.join(workspace.repo_dir, "local.cfg")
            self.render_config(local_cfg_src, local_cfg, tmpl_vars)
            configs += (local_cfg,)

        # Check the config files are valid YAML or JSON
        self._validate_configs(configs)

        if cli:
            cli_args = ()
            for config in configs:
                cli_args += ('-c', config)
            return cli_args
        else:
            return configs

    def show(self, project, workspace, series, stage):
        """
        Display rendered juju-deployer configuration files
        """
        # Quiet logging to get clean display of YAML
        logger = logging.getLogger()
        logger.setLevel(level='WARNING')
        configs = self._deployer_configs(project, workspace, series, stage,
                                         cli=False)
        deploy_name = self._get_target(configs)
        deployer_config = ConfigStack(configs)
        deployer_config.load()
        deployment = deployer_config.get(deploy_name)
        # XXX Pretty much a gross hack, but
        #     Config.get() forces our hand.
        deployment.repo_path = workspace.repo_dir
        deployment.resolve_config()
        print(yaml.dump(deployment.data))

    def diff(self, project, workspace, series, stage):
        """
        Display juju-deployer --diff comparing running environment
        to the juju-deployer configuration
        """
        deployer = get_dep_prog("juju-deployer")
        configs = self._deployer_configs(project, workspace, series, stage)
        deploy_name = self._get_target(configs)
        cmd = ('python', deployer)
        cmd += configs
        cmd += ('--diff', deploy_name)
        with chdir(workspace.repo_dir):
            subprocess.call(cmd)

    def run(self, project, workspace, series, stage, auto_secrets=True):
        logging.debug("### Running phase {} with options {}"
                      "".format(self.name, self.options))
        logging.info("Running juju-deployer")
        deployer = get_dep_prog("juju-deployer")

        configs = self._deployer_configs(project, workspace, series, stage,
                                         auto_secrets=auto_secrets)

        juju_env = None
        if "juju" in self.options:
            juju_env = self.options['juju']
            configs += ('-e', juju_env)
        elif os.environ.get("JUJU_ENV"):
            juju_env = os.environ.get("JUJU_ENV")
            configs += ('-e', juju_env)

        if "delay" in self.options:
            deploy_delay = ('-s', str(self.options.get("delay")))
        else:
            deploy_delay = ('-s', '0')

        status_timeout = str(self.options.get("status-timeout", "1800"))

        if "timeout" in self.options:
            timeout = ('-t', str(self.options.get("timeout")))
        else:
            timeout = ('-t', '2700')

        if "retry" in self.options:
            retry = ('-r', str(self.options.get("retry")))
        else:
            retry = ()

        if "wait" in self.options:
            wait_flag = str2bool(self.options['wait'])
        else:
            wait_flag = False

        if "optional" in self.options:
            optional = str2bool(self.options['optional'])
        else:
            optional = False

        deploy_name = self._get_target(configs)

        # run the deploy
        env = os.environ.copy()
        env['MOJO_SERIES'] = series
        env['MOJO_STAGE'] = stage or 'devel'
        cmd = ('python', deployer)
        cmd += configs
        cmd += deploy_delay
        cmd += timeout
        cmd += retry
        cmd += ('-d', deploy_name, '-W', '-u')

        logging.info("Running: {}".format(" ".join(cmd)))

        if optional:
            logging.info("NOTE: This phase is marked optional; "
                         "failure will not terminate the run.")

        with chdir(workspace.repo_dir):
            status, output = bicommand(cmd, showoutput=True, env=env)
            logging.getLogger('file').info(output)

        if status != 0:
            timeout_fail = re.search("Reached deployment timeout.. exiting",
                                output)
            if timeout_fail:
                logging.error("Deployment timed out. It was set to {} "
                              "seconds".format(timeout[1]))
            hook_fail = re.search(".*unit: (?P<unit>[^:]+):.*hook failed",
                             output)
            if hook_fail:
                failed_unit = hook_fail.group("unit")
                logging.error("Deployment failed, grabbing the last 200 lines of "
                              "output from the juju logs on {}".format(failed_unit))
                cmd = ["juju", "ssh", failed_unit, "sudo", "tail", "-200",
                       "/var/log/juju/unit-%s.log" % (
                            failed_unit.replace("/", "-"),)]
                with chdir(workspace.repo_dir):
                    failure = subprocess.check_output(cmd)
                    logging.error(failure)
            # Show juju status
            juju_status = mojo.juju.status.Status(environment=juju_env)
            logging.error("Juju Status: {}".format(juju_status.status()))
            if not timeout_fail and not hook_fail:
                logging.error("There was an unrecognised problem with "
                              "running a deploy phase")

            if optional:
                logging.info("Error found during optional deployment phase: continuing as directed.")
            else:
                raise JujuDeployerException("Error found during deployment phase")
        else:
            logging.info("Checking Juju status")
            juju_status = mojo.juju.status.Status(environment=juju_env)
            juju_status.check_and_wait(timeout=status_timeout)
            if wait_flag:
                logging.info("Waiting for environment to reach steady state")
                try:
                    wait()
                except JujuWaitException:
                    # Show juju status
                    logging.error("Juju Status: {}".format(print_juju_status(environment=juju_env)))
                    raise
                logging.info("Environment has reached steady state")

    def render_config(self, src, dest, tmpl_vars=None):
        if not tmpl_vars:
            tmpl_vars = {}
        with open(src) as inp:
            with open(dest, "w") as outp:
                try:
                    outp.write(Template(inp.read()).render(**tmpl_vars))
                except TemplateSyntaxError as e:
                    raise InvalidYAMLException("Templating Error due to "
                                               "invalid YAML or JSON in "
                                               "juju-deployer configuration "
                                               "file {}. Likely a missing end "
                                               "curly bracket:\n{}".format(src,
                                                                           e))

        # Rendered deploy configs may contain secrets
        os.chmod(dest, 0600)


def int_from_dict(d, k, default):
        try:
            return int(d[k])
        except KeyError:
            pass
        except ValueError:
            logging.warning(
                "Invalid value {} for '{}', "
                "using {} instead".format(
                    d[k], k, default))

        return default


class VerifyPhase(ScriptPhase):
    """Execute a script to verify environment"
       Arguments:
           config      base name for script file
           lxc         Boolean to use lxc or not
           retry       int; how many times to retry
           sleep       int; how many seconds to wait between retries
    """
    name = "verify"

    def run(self, project, workspace, series, stage, auto_secrets=True):
        retry = int_from_dict(self.options, 'retry', 1)
        sleep = int_from_dict(self.options, 'sleep', 5)

        _run = super(VerifyPhase, self).run

        for i in xrange(1, retry + 1):
            if i > 1:
                logging.info(
                    "Retrying (attempt {} of {})".format(
                        i, retry))

            try:
                _run(project, workspace, series, stage,
                     auto_secrets=auto_secrets)
            except ScriptPhaseException as e:
                if i < retry:
                    logging.warning(e.output)
                    logging.warning(
                        "Error: verify script exited with status {} "
                        "(attempt {} of {}); sleeping for {}s".format(
                            e.returncode, i, retry, sleep))
                    time.sleep(sleep)
                else:
                    raise VerifyPhaseException(e.output)
            else:
                # Hooray, success!
                break


class StopOnPhase(ScriptPhase):
    """Execute a script that checks if we want to stop further execution on a
       specific return code.
       Arguments:
            config          base name for the script file
            return_code     the return code to stop further execution on
    """
    name = "stop-on"

    def run(self, project, workspace, series, stage):

        if "return-code" in self.options:
            return_code = int(self.options['return-code'])
            if return_code < 1 or return_code > 255:
                raise InvalidStopOnReturnCodeException("Invalid return code, must be between 1 and 255")
        else:
            raise InvalidStopOnReturnCodeException("Invalid return code, must be between 1 and 255")

        _run = super(StopOnPhase, self).run

        try:
            _run(project, workspace, series, stage)
        except ScriptPhaseException as e:
            if e.returncode == return_code:
                # Exit silently
                logging.info(e.output)
                sys.exit(0)
            else:
                raise StopOnPhaseException(e.output)


class JujuCheckWaitPhase(Phase):
    """Check juju status and then confirm it's in a steady state using 
    lp:juju-wait
    
        Arguments:
            status-timeout  We check juju status after each deploy phase. How
                            long in seconds before we timeout if instances remain
                            in a pending status.
                            Default: 1800 seconds (30 minutes)
    """
    name = "juju-check-wait"

    def run(self, project, workspace, series, stage=None):
        status_timeout = str(self.options.get("status-timeout", "1800"))
        logging.info("Checking Juju status")
        juju_status = mojo.juju.status.Status()
        juju_status.check_and_wait(timeout=status_timeout)
        logging.info("Waiting for environment to reach steady state")
        try:
            wait()
        except JujuWaitException:
            # Show juju status
            logging.error("Juju Status: {}".format(print_juju_status()))
            raise
        logging.info("Environment has reached steady state")


# ========================================
# Phase creation helper
# ========================================

def phase_types(phase_class=None):
    "Build a dict of phases and their subtypes"
    if phase_class is None:
        phase_class = Phase
    types = {phase_class.name: phase_class}
    for subclass in phase_class.__subclasses__():
        types.update(phase_types(subclass))
    return types


def build(phase_name, **kwargs):
    "Build a phase by name, passing all arguments as a single string"
    if getattr(phase_types, "cache", None) is None:
        phase_types.cache = phase_types(Phase)
    if phase_name not in phase_types.cache:
        raise NoSuchPhaseError("{} is not a registered phase type"
                               "".format(phase_name))
    return phase_types.cache[phase_name](**kwargs)


def str2bool(v):
    if v:
        return v.lower() in ("yes", "true", "t", "1")
    return False
