import logging
import os

from mojo import bzr
from mojo.exceptions import SpecUninitializedError, ConfigNotFoundException
from mojo.manifest import Manifest


class Spec(object):
    def __init__(self, spec_dir, remote=None, stage=None, manifest_file=None):
        self.spec_dir = spec_dir
        self.manifest_file = manifest_file
        self.local_branch = bzr.BzrBranch(self.spec_dir)
        if remote:
            self.remote_branch = bzr.BzrBranch(remote)
        elif self.is_initialized:
            if self.local_branch.parent:
                self.remote_branch = bzr.BzrBranch(self.local_branch.parent)
            else:
                self.remote_branch = None
        else:
            raise ValueError("No remote or local branch specified for Spec")
        self.stage = stage
        self._manifest = None

    @property
    def is_initialized(self):
        return self.local_branch is not None and self.local_branch.is_branch

    def initialize(self, clean_spec=True):
        "Download or update spec on disk"
        if not os.path.isdir(self.spec_dir):
            os.makedirs(self.spec_dir)
        if self.local_branch.is_branch:
            self.local_branch.pull(src_location=self.remote_branch.location,
                                   overwrite=clean_spec)
        else:
            self.local_branch = self.remote_branch.clone(self.spec_dir)
        if clean_spec:
            self.local_branch.revert()
            self.local_branch.cleantree()

    @property
    def manifest(self):
        logging.info("Retrieve the spec's manifest")
        if self._manifest:
            return self._manifest
        if not self.is_initialized:
            return None
        self._manifest = Manifest(self, manifest_file=self.manifest_file,
                                  stage=self.stage)
        return self._manifest

    def _remove_duplicates(self, mylist):
        """Remove duplicates while preserving order.

        See http://stackoverflow.com/a/480227/523729
        """
        seen = set()
        seen_add = seen.add
        return [
            x for x in mylist
            if not (x in seen or seen_add(x))]

    def iterate_subpaths(self, path):
        """Iterate subpaths of a path, from most to least explicit.
        e.g. input of 'a/b/c' will yield ('a/b/c', 'b/c', 'c')"""
        if path:
            yield path
            parts = path.split('/', 1)
            if len(parts) == 1:
                yield parts[0]
            else:
                for _path in self.iterate_subpaths(parts[1]):
                    yield _path

    def get_configs(self, base, stage=None):
        """Locate a list of stage-qualified configuration files.
           Configs are returned relative to the spec dir from
           least specific to most specfic in this order:
             * <base>
             * <stage>/../<base>
             * <stage>/<base>"""
        if not self.is_initialized:
            raise SpecUninitializedError("Spec has no local copy: {}"
                                         "".format(self.spec_dir))
        config_names = []
        if stage:
            # Common edge case $MOJO_STAGE/../$CONFIG
            config_names.append(os.path.join(stage, '..', base))
            for path in self.iterate_subpaths(os.path.join(stage, base)):
                config_names.append(path)
        else:
            for path in self.iterate_subpaths(base):
                config_names.append(path)

        config_names = [os.path.normpath(config) for config in config_names]
        config_names = self._remove_duplicates(config_names)
        # Order is important here
        # Least specific to most specific
        # Sort by string length
        config_names.sort(key=len)

        configs = [
            config for config in config_names
            if os.path.isfile(os.path.join(self.spec_dir, config))]

        if configs:
            return configs

        raise ConfigNotFoundException(
            "Config file '{}' not found.\n"
            "Is the MOJO_STAGE environment variable or the --stage option "
            "set?\nIs {} a spec branch?".format(base,
                                                self.local_branch.location),
            config_names)

    def get_config(self, base, stage=None):
        """Locate a stage-qualified configuration file.
           The absolute path to the most specific match,
           the first of the following configs is returned:
             * <stage>/<base>
             * <stage>/../<base>
             * <base>"""
        # The last element is the most specific matching path
        return os.path.join(self.spec_dir, self.get_configs(base, stage)[-1])


