# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import os
import sys
from unittest import TestCase
# Make sure that the local mojo is importable
local_copy = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "..")
if os.path.exists(local_copy) and os.path.isdir(local_copy):
    sys.path.insert(0, local_copy)

import mojo
from bzrlib.branch import (
    Branch,
    BzrBranch7,
    BzrBranch8,
)
from bzrlib.workingtree import WorkingTree
from tempfile import mkdtemp, mkstemp
from shutil import rmtree
import subprocess
from datetime import datetime


def init_branch(directory):
    try:
        cmd = ['bzr', 'init', directory]
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError:
        raise


def checkout_branch(branch, dest, revno=None, verbose=False, tree=True):
    cmd = ['bzr', 'branch']
    if revno is not None:
        cmd.append("-r{}".format(revno))
    if not tree:
        cmd.append("--no-tree")
    if os.path.isdir(dest) and not os.listdir(dest):
        # empty directory
        cmd.append("--use-existing-dir")
    cmd.extend((branch, dest))
    try:
        if verbose:
            subprocess.check_call(cmd)
        else:
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError:
        raise
    return dest


def create_unique_branch(directory, filenames=None, manifests=None):
    init_branch(directory)
    if filenames is None:
        filenames = ('test_file',)
    for filename in filenames:
        filepath = os.path.join(directory, filename)
        if not os.path.exists(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))
        with open(filepath, 'w') as f:
            f.write('deploy-config:')
            f.write('  ' + datetime.now().isoformat())
            f.flush()
    # Ensure there's a valid manifest as part of the spec.
    with open(os.path.join(directory, 'manifest'), 'w') as f:
        f.write("\n".join(["collect", "repo", "deploy"]))

    manifests = manifests or {'manifest': 'collect\nrepo\ndeploy'}
    for manifest_path, content in manifests.items():
        parent, _ = os.path.split(manifest_path)
        full_parent = os.path.join(directory, parent)
        if not os.path.exists(full_parent):
            os.makedirs(full_parent)
        with open(os.path.join(directory, manifest_path), 'w+') as fp:
            fp.write(content)

    filenames = list(filenames) + manifests.keys()

    with mojo.utils.chdir(directory):
        for filename in filenames:
            cmd = ['bzr', 'add', filename]
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        cmd = ['bzr', 'commit', '-m', 'commit']
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)


def dirUrl(path):
    if not os.path.isabs(path):
        path = os.path.abspath(path)
    parts = ["file://", path]
    if path[-1] != "/":
        parts.append("/")
    return "".join(parts)


class BzrBranchTest(TestCase):
    # this lp URL must be a branch that exists
    lp_url = "http://bazaar.launchpad.net/~mew/+junk/test-target/"
    lp_url_bad = "http://bazaar.launchpad.net/~INVALID/INVALID"
    lcl_br_bad = "/INVALID/PATH"

    def setUp(self):
        super(BzrBranchTest, self).setUp()
        self.uniq_br = self.mkTempDir()
        create_unique_branch(self.uniq_br)

        self.addCleanup(os.chdir, os.getcwd())

    def mkTempDir(self):
        tmpdir = mkdtemp()
        self.addCleanup(rmtree, tmpdir)
        return tmpdir

    def test_normalizes_local_paths(self):
        location = 'relative/path'
        abs_location = os.path.normpath(os.path.join(os.getcwd(), location))

        br = mojo.bzr.BzrBranch(location)
        self.assertTrue(os.path.isabs(br.location))
        self.assertEqual(br.location, abs_location)

    def test_gets_branch(self):
        for location in [self.lp_url, self.uniq_br, dirUrl(self.uniq_br)]:
            br = mojo.bzr.BzrBranch(location)
            self.assertIsInstance(br.bzr_branch,
                                  (Branch, BzrBranch7, BzrBranch8))

    def test_is_branch(self):
        for location in [self.lp_url, self.uniq_br, dirUrl(self.uniq_br)]:
            br = mojo.bzr.BzrBranch(location)
            self.assertTrue(br.is_branch)

        for location in [self.lp_url_bad, self.lcl_br_bad]:
            br = mojo.bzr.BzrBranch(location)
            self.assertFalse(br.is_branch)

    def test_gets_tree(self):
        br = mojo.bzr.BzrBranch(self.uniq_br)
        self.assertIsInstance(br.bzr_tree, (WorkingTree))

        treeless = checkout_branch(self.uniq_br, self.mkTempDir(), tree=False)
        br = mojo.bzr.BzrBranch(treeless)
        self.assertIsNone(br.bzr_tree)

        remote = mojo.bzr.BzrBranch(self.lp_url)
        self.assertIsNone(remote.bzr_tree)

    def test_has_tree(self):
        br = mojo.bzr.BzrBranch(self.uniq_br)
        self.assertTrue(br.has_tree)

        tr_loc = checkout_branch(self.uniq_br, self.mkTempDir(), tree=False)
        br = mojo.bzr.BzrBranch(tr_loc)
        self.assertFalse(br.has_tree)

    def test_mangles_file_urls(self):
        br_url = dirUrl(self.uniq_br)
        br = mojo.bzr.BzrBranch(br_url)
        self.assertEqual(br.location, br_url[7:])

    def test_revno(self):
        br = mojo.bzr.BzrBranch(self.uniq_br)
        self.assertIsInstance(br.revno, int)
        self.assertEqual(br.revno, 1)

    def test_revno_tree(self):
        src_br = mojo.bzr.BzrBranch(self.uniq_br)
        dest_loc = self.mkTempDir()
        dest_br = src_br.clone(dest_loc, 0)
        self.assertIsInstance(dest_br.revno_tree, int)
        self.assertEqual(dest_br.revno_tree, 0)

        remote_br = mojo.bzr.BzrBranch(self.lp_url)
        self.assertIsNone(remote_br.revno_tree)

    def test_clones_to_dir(self):
        src_br = mojo.bzr.BzrBranch(self.uniq_br)
        dest_loc = self.mkTempDir()
        clone = src_br.clone(dest_loc)
        self.assertIsInstance(clone, mojo.bzr.BzrBranch)
        self.assertTrue(clone.is_branch)
        self.assertEqual(clone.parent, dirUrl(self.uniq_br))
        self.assertEqual(src_br.revno, clone.revno)

        dest_loc = self.mkTempDir()
        clone = src_br.clone(dest_loc, 0)
        self.assertEqual(clone.revno, 0)

    def test_clone_nothing(self):
        empty_dir = self.mkTempDir()
        br = mojo.bzr.BzrBranch(empty_dir)
        self.assertIsNone(br.clone('foo'))

    def test_gets_parent(self):
        src_br = mojo.bzr.BzrBranch(self.uniq_br)

        dest_loc = self.mkTempDir()
        dest_br = src_br.clone(dest_loc)
        self.assertEqual(dest_br.parent, dirUrl(src_br.location))

        remote_br = mojo.bzr.BzrBranch(self.lp_url)
        self.assertIsNone(remote_br.parent)

        empty_br = mojo.bzr.BzrBranch(self.mkTempDir())
        self.assertIsNone(empty_br.parent)

    def test_detect_and_revert_changes(self):
        br = mojo.bzr.BzrBranch(self.uniq_br)
        self.assertFalse(br.has_changes())

        c_file = os.path.join(self.uniq_br, os.listdir(self.uniq_br)[0])
        with open(c_file, "w") as f:
            f.write("test_has_changes")
            f.flush()
            f.close()
        self.assertTrue(br.has_changes())

        br.revert()
        self.assertFalse(br.has_changes())
        self.assertTrue(br.has_changes(detritus=True))

        new_dir = mkdtemp(dir=self.uniq_br)
        fd, new_file_path = mkstemp(dir=new_dir)
        new_file = os.fdopen(fd, "w")
        new_file.write("some text")
        new_file.flush()
        new_file.close()
        self.assertFalse(br.has_changes())
        self.assertTrue(br.has_changes(unknowns=True))

        br.cleantree()
        self.assertFalse(br.has_changes())

        remote_br = mojo.bzr.BzrBranch(self.lp_url)
        self.assertFalse(remote_br.has_changes(unknowns=True, detritus=True))

        tr_loc = checkout_branch(br.location, self.mkTempDir(), tree=False)
        treeless = mojo.bzr.BzrBranch(tr_loc)
        self.assertFalse(treeless.has_changes(unknowns=True, detritus=True))
        with self.assertRaises(mojo.bzr.NoWorkingTreeError):
            treeless.revert()
        with self.assertRaises(mojo.bzr.NoWorkingTreeError):
            treeless.cleantree()

        with self.assertRaises(mojo.bzr.NoWorkingTreeError):
            mojo.bzr.BzrBranch(self.lp_url).revert()
        with self.assertRaises(mojo.bzr.NoWorkingTreeError):
            mojo.bzr.BzrBranch(self.lp_url).cleantree()

    def test_creates_repo(self):
        br = mojo.bzr.BzrBranch(self.mkTempDir())
        br.create()
        self.assertTrue(br.is_branch)

        with self.assertRaises(mojo.bzr.AlreadyControlDirError):
            br.create()

    def test_pull_updates_existing(self):
        br_loc = checkout_branch(self.uniq_br, self.mkTempDir(), 0)
        br = mojo.bzr.BzrBranch(br_loc)
        self.assertEqual(br.revno, 0)
        self.assertTrue(br.pull())
        self.assertNotEqual(br.revno, 0)

        br_loc = checkout_branch(self.uniq_br, self.mkTempDir(), 0, tree=False)
        br = mojo.bzr.BzrBranch(br_loc)
        self.assertEqual(br.revno, 0)
        self.assertTrue(br.pull())
        self.assertNotEqual(br.revno, 0)

    def test_pull_refuses_update_of_nothing(self):
        br = mojo.bzr.BzrBranch(self.mkTempDir())
        with self.assertRaises(mojo.bzr.NotABranchError):
            br.pull()
        with self.assertRaises(mojo.bzr.NotABranchError):
            br.pull(src_location=self.lp_url)

    def test_pull_skips_update_from_nothing(self):
        br = mojo.bzr.BzrBranch(self.uniq_br)
        with self.assertRaises(mojo.bzr.NotABranchError):
            br.pull(src_location=self.lp_url_bad)

    def test_pushes(self):
        src_loc = self.mkTempDir()
        create_unique_branch(src_loc)
        src_br = mojo.bzr.BzrBranch(src_loc)

        # push to non-repo, returns None
        dest_loc = self.mkTempDir()
        dest_br = src_br.push(dest_loc)
        self.assertIsNone(dest_br)

        # push to same repo, returns target branch
        dest_loc = self.mkTempDir()
        src_br.clone(dest_loc, 0)
        dest_br = src_br.push(dest_loc)
        self.assertIsInstance(dest_br, mojo.bzr.BzrBranch)
        self.assertNotEqual(dest_br.revno, 0)

        # push to non-ancestor, exception
        with self.assertRaises(mojo.bzr.DivergedBranchError):
            dest_br = src_br.push(self.uniq_br)

        # overwrite non-ancestor, returns target branch
        dest_loc = self.mkTempDir()
        create_unique_branch(dest_loc)
        dest_br = src_br.push(dest_loc, overwrite=True)
        self.assertIsInstance(dest_br, mojo.bzr.BzrBranch)
