# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
from unittest import TestCase

from tempfile import mkdtemp
from shutil import rmtree
import commands
import os
import sys
import mock
import json
import mojo.project
import mojo.spec

from mojo.utils import chdir
from mojo.tests.utils import MojoTestCaseMixin

# Make sure that the local mojo is importable
import mojo.workspace

local_copy = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "..",
    "mojo")
if os.path.exists(local_copy) and os.path.isdir(local_copy):
    sys.path.insert(0, local_copy)

try:
    import mojo
except ImportError:
    raise ImportError("Could not import local copy of mojo.")


class WorkspaceTest(TestCase):

    def setUp(self):
        super(WorkspaceTest, self).setUp()
        self.bad_path = "/INVALID/PATH"
        self.good_path = mkdtemp()
        self.addCleanup(rmtree, self.good_path)

        self.build_dir_name = "build"
        self.local_dir_name = "local"
        self.log_dir_name = "log"
        self.spec_dir_name = "spec"
        self.all_names = [
            self.build_dir_name,
            self.local_dir_name,
            self.log_dir_name,
            self.spec_dir_name,
        ]

    def test_workspace_sets_path(self):
        w = mojo.workspace.Workspace("workspace_name", self.good_path)
        self.assertEqual(w.path, self.good_path)

        with self.assertRaises(mojo.workspace.NoSuchWorkspaceError):
            w = mojo.workspace.Workspace("workspace_name", self.bad_path)

    def test_workspace_sets_default_path(self):
        with chdir(self.good_path):
            w = mojo.workspace.Workspace("workspace_name", None)
        self.assertEqual(w.path, self.good_path)

    def test_prepares_path(self):
        w = mojo.workspace.Workspace("workspace_name", self.good_path)
        self.assertTrue(w.prepare())

        all_subdirs = [os.path.join(self.good_path, d) for d in self.all_names]
        for subdir in all_subdirs:
            self.assertTrue(os.path.exists(subdir))

    def test_prepares_new_path(self):
        w = mojo.workspace.Workspace("workspace_name", self.good_path)
        self.assertTrue(w.prepare())

        all_subdirs = [os.path.join(self.good_path, d) for d in self.all_names]
        for subdir in all_subdirs:
            self.assertTrue(os.path.exists(subdir))

    def test_prepare_errors_return_false(self):
        os.chmod(self.good_path, 0o555)
        w = mojo.workspace.Workspace("workspace_name", self.good_path)
        self.assertFalse(w.prepare())

    def test_returns_dirs(self):
        w = mojo.workspace.Workspace("workspace_name", self.good_path)

        build_dir = os.path.join(self.good_path, self.build_dir_name)
        self.assertEqual(w.build_dir, build_dir)

        local_dir = os.path.join(self.good_path, self.local_dir_name)
        self.assertEqual(w.local_dir, local_dir)

        log_dir = os.path.join(self.good_path, self.log_dir_name)
        self.assertEqual(w.log_dir, log_dir)

        spec_dir = os.path.join(self.good_path, self.spec_dir_name)
        self.assertEqual(w.spec_dir, spec_dir)


class SpecTest(TestCase, MojoTestCaseMixin):

    def test_inits_branch(self):
        spec_dir = self.create_spec_dir(configs=['cf', 'cf-env'])
        spec = mojo.spec.Spec(spec_dir)

        self.assertIsInstance(spec.local_branch, mojo.bzr.BzrBranch)
        self.assertEqual(spec.local_branch.location, spec_dir)

    def test_get_config(self):
        cases = [{
            "config_files": ['staging/service.yml'],
            "input": dict(base="service.yml", stage='staging'),
            "expected": "staging/service.yml",
        }, {
            "config_files": ['service.yml', 'staging/service.yml'],
            "input": dict(base="service.yml", stage='staging'),
            "expected": "staging/service.yml",
        }, {
            "config_files": ['service.yml', 'service.yml-staging'],
            "input": dict(base="service.yml", stage=None),
            "expected": "service.yml",
        }, {
            "config_files": [
                'service.yml',
                'pes/service.yml',
                'pes/staging/service.yml',
            ],
            "input": dict(base="service.yml", stage='pes/staging'),
            "expected": "pes/staging/service.yml",
        }, {
            "config_files": ['service.yml', 'pes/service.yml'],
            "input": dict(base="service.yml", stage='pes/staging'),
            "expected": "pes/service.yml",
        }]
        for test_case in cases:
            spec_dir = self.create_spec_dir(test_case['config_files'],
                                            stage=test_case['input']['stage'])

            spec = mojo.spec.Spec(spec_dir)

            expected_config = os.path.join(spec_dir, test_case["expected"])
            actual_config = spec.get_config(**test_case['input'])
            self.assertEqual(expected_config, actual_config)

    def test_get_configs(self):
        cases = [{
            "config_files": ['staging/service.yml'],
            "input": dict(base="service.yml", stage='staging'),
            "expected": ["staging/service.yml"],
        }, {
            "config_files": ['service.yml', 'staging/service.yml'],
            "input": dict(base="service.yml", stage='staging'),
            "expected": ["service.yml", "staging/service.yml"],
        }, {
            "config_files": [
                'staging/service.yml',
                'service.yml',
            ],
            "input": dict(base="service.yml", stage='staging'),
            "expected": [
                "service.yml",
                'staging/service.yml',
            ],
        }, {
            "config_files": ['service.yml', 'service.yml-staging'],
            "input": dict(base="service.yml", stage=None),
            "expected": ["service.yml"],
        }, {
            "config_files": [
                'service.yml',
                'pes/service.yml',
                'pes/staging/service.yml',
            ],
            "input": dict(base="service.yml", stage='pes/staging'),
            "expected": [
                "service.yml",
                "pes/service.yml",
                "pes/staging/service.yml",
            ],
        }, {
            "config_files": ['service.yml', 'pes/service.yml'],
            "input": dict(base="service.yml", stage='pes/staging'),
            "expected": ["service.yml", "pes/service.yml"],
        }]
        for test_case in cases:
            spec_dir = self.create_spec_dir(test_case['config_files'],
                                            stage=test_case['input']['stage'])

            spec = mojo.spec.Spec(spec_dir)

            actual_configs = spec.get_configs(**test_case['input'])
            self.assertEqual(test_case["expected"], actual_configs)


class BicommandTest(TestCase):
    "Test that bicommand produces the same output as getstatusoutput"
    def test_bicommand(self):
        cmds = [
            ["/bin/ls", "-1", "/tmp"],  # rarely will fail, will usually be fast
            ["/bin/ls", "/usr"],  # never (?) will fall, will usually be fast
            ["/bin/uname"],  # shouldn't ever fail, will be fast
            ["/usr/bin/find", "/this/file/should/not/exist"],  # should be stderr
            ["/usr/bin/find", "/usr/share/doc/dpkg"],  # will be slow
        ]
        for cmd in cmds:
            bc_status, bc_output = mojo.utils.bicommand(cmd)
            c_status, c_output = commands.getstatusoutput(' '.join(cmd))

            # We don't care about return codes, we just want to confirm they're
            # both returning zero or non-zero
            if bc_status != 0:
                bc_status = 1
            if c_status != 0:
                c_status = 1

            self.assertEqual(bc_status, c_status)
            self.assertEqual(bc_output, c_output)


class ProjectTest(TestCase):

    def test_project_lxc_trusty(self):
        project = mojo.project.Project('my-project', 'trusty', '/srv/mojo', container_class=mojo.contain.LXCContainer)
        self.assertTrue(type(project.container) == mojo.contain.LXCContainer)
        self.assertTrue(project.container.container_root == '/srv/mojo/my-project/trusty/ROOTFS')

    def test_project_lxc_trusty_exists(self):
        """Pre-existing project has lxc root at
        /var/lib/lxc/my-project.trusty/rootfs"""
        mojo_root = mkdtemp()
        self.addCleanup(rmtree, mojo_root)
        project_series_dir = os.path.join(mojo_root, 'my-project', 'trusty')
        mojo.utils.mkdirs(project_series_dir)
        project_info_file = os.path.join(mojo_root, 'my-project', 'trusty', '.project')
        existing_project_info = {'container_root': '/var/lib/lxc/my-project.trusty/rootfs',
                                 'container_class': 'lxc'}
        with open(project_info_file, 'w') as info:
            info.write(json.dumps(existing_project_info))

        project = mojo.project.Project('my-project', 'trusty', mojo_root, container_class=mojo.contain.LXCContainer)
        self.assertTrue(type(project.container) == mojo.contain.LXCContainer)
        self.assertTrue(project.container.container_root == '/var/lib/lxc/my-project.trusty/rootfs')

    @mock.patch('mojo.contain.LXCContainer.lxc_version', return_value='0.7.5')
    def test_project_lxc_precise(self, _lxc_version):
        mojo.contain.apt_pkg.init()
        mojo.contain.apt_pkg.Cache()
        project = mojo.project.Project('my-project', 'precise', '/srv/mojo', container_class=mojo.contain.LXCContainer)
        self.assertTrue(type(project.container) == mojo.contain.LXCContainer)
        self.assertTrue(project.container.container_root == '/var/lib/lxc/my-project.precise/rootfs')

    def test_project_containerless(self):
        project = mojo.project.Project('my-project', 'trusty', '/srv/mojo', container_class=mojo.contain.NoContainer)
        self.assertTrue(type(project.container) == mojo.contain.NoContainer)
        self.assertTrue(project.container.container_root == '/srv/mojo/my-project/trusty')

    def test_project_containerless_exists(self):
        """Pre-existing containerless project"""
        mojo_root = mkdtemp()
        self.addCleanup(rmtree, mojo_root)
        project_series_dir = os.path.join(mojo_root, 'my-project', 'trusty')
        mojo.utils.mkdirs(project_series_dir)
        project_info_file = os.path.join(project_series_dir, '.project')
        existing_project_info = {'container_root': project_series_dir,
                                 'container_class': 'containerless'}
        with open(project_info_file, 'w') as info:
            info.write(json.dumps(existing_project_info))

        # Recieves the default LXCContainer but is a pre-existing
        # NoContainer project
        project = mojo.project.Project('my-project', 'trusty', mojo_root, container_class=mojo.contain.LXCContainer)
        self.assertTrue(type(project.container) == mojo.contain.NoContainer)
        self.assertTrue(project.container.container_root == project_series_dir)
