# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import mock
import os
import stat
from unittest import TestCase

from mojo.phase import DeployerPhase, CollectPhase, CollectPhaseException, CharmRepoPhase
from mojo.tests import utils
from mojo.project import Project


class DeployerPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    def get_deployer_command_for_run(self, phase, workspace, stage='staging'):
        """Return the deployer command executed when the phase runs."""
        # XXX: It'd be nicer to stub mojo.utils.subprocess.Popen
        # but bicommand makes that quite complicated.
        with mock.patch('mojo.phase.bicommand') as mock_bicommand:
            with mock.patch('mojo.phase.wait'):
                with mock.patch('mojo.juju.status.Status.check_and_wait'):
                    mock_bicommand.return_value = (0, 'fake-output')

                    project = Project('unused', 'trusty', '/srv/mojo')
                    phase.run(project, workspace, 'trusty', stage)

                    return mock_bicommand.call_args[0][0]

    def test_multiple_configs_rendered_for_deployer(self):
        phase = DeployerPhase(config='services.cfg')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/services.cfg', 'services.cfg'],
            stage='staging')

        command = self.get_deployer_command_for_run(phase, workspace)

        # Verify that the configs were both rendered.
        self.assertTrue(os.path.exists(
            '{}/services.cfg'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/services.cfg'.format(workspace.repo_dir)))
        # Verify that the configs were included in the deployer command.
        expected_configs = (
            '-c', '{configs_dir}/services.cfg'.format(
                configs_dir=workspace.repo_dir),
            '-c', '{configs_dir}/staging/services.cfg'.format(
                configs_dir=workspace.repo_dir),
        )
        self.assertSubsequenceInSequence(expected_configs, command)

    def test_local_config_rendered_with_same_perms(self):
        phase = DeployerPhase(config='services', local="local.yaml")
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['services'],
            stage='staging')
        utils.touch(os.path.join(workspace.local_dir, "local.yaml"))

        command = self.get_deployer_command_for_run(phase, workspace)

        # The local.yaml is rendered to $MOJO_REPO_DIR/local.cfg (?)
        expected_configs = (
            '-c', '{configs_dir}/services'.format(
                configs_dir=workspace.repo_dir),
            '-c', '{configs_dir}/local.cfg'.format(
                configs_dir=workspace.repo_dir),
        )
        self.assertSubsequenceInSequence(expected_configs, command)
        mode = os.stat(os.path.join(workspace.repo_dir, 'local.cfg')).st_mode
        group_other_bits = (stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH |
                            stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP)

        self.assertFalse(mode & group_other_bits,
                         "The rendered local config is not accessible to "
                         "group or others.")

    @mock.patch('mojo.phase.SecretsPhase')
    def test_secrets_automatically_run(self, secrets):
        phase = DeployerPhase(config='services.cfg', target='test')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/services.cfg', 'services.cfg'],
            stage='staging')

        self.get_deployer_command_for_run(phase, workspace)

        secrets.assert_called_with(auto_secrets=True)
        called_methods = [c[0] for c in secrets.mock_calls]
        assert ('().run' in called_methods)


class CollectPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    @mock.patch('codetree.config.Config')
    def test_codetree_is_called(self, codetree_Config):
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        phase.run('unused', workspace, 'trusty', 'staging')

        codetree_Config.assert_called_with(os.path.join(workspace.spec_dir,
                                           'collect'))
        called_methods = [c[0] for c in codetree_Config.mock_calls]
        assert ('().build' in called_methods)

    @mock.patch('codetree.config.Config')
    def test_codetree_failure_exception(self, _codetree_Config):
        _codetree_Config.return_value.build.return_value = False
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        with self.assertRaises(CollectPhaseException):
            phase.run('unused', workspace, 'trusty', 'staging')

    @mock.patch('mojo.phase.CharmRepoPhase')
    @mock.patch('codetree.config.Config')
    def test_repo_automatically_runs(self, codetree_Config, charm_repo):
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        phase.run('unused', workspace, 'trusty', 'staging')

        called_methods = [c[0] for c in charm_repo.mock_calls]
        assert ('().run' in called_methods)


class CharmRepoPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    @mock.patch('mojo.charm_repo.CharmRepo.build_from_config')
    @mock.patch('mojo.charm_repo.CharmRepo.build_all')
    def test_repo_default(self, build_all, build_from_config):
        phase = CharmRepoPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['repo', 'services.cfg'], stage='staging')

        phase.run('unused', workspace, 'trusty', 'staging')

        build_from_config.assert_called_with(
            os.path.join(workspace.spec_dir, 'repo'),
            os.path.join(
                os.path.dirname(workspace.spec_dir), 'build'),
            'trusty')

    @mock.patch('mojo.charm_repo.CharmRepo.build_from_config')
    @mock.patch('mojo.charm_repo.CharmRepo.build_all')
    def test_repo_with_config(self, build_all, build_from_config):
        phase = CharmRepoPhase(config='repo-update')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['repo-update', 'services.cfg'], stage='staging')

        phase.run('unused', workspace, 'trusty', 'staging')

        build_from_config.assert_called_with(
            os.path.join(workspace.spec_dir, 'repo-update'),
            os.path.join(
                os.path.dirname(workspace.spec_dir), 'build'),
            'trusty')

    @mock.patch('mojo.charm_repo.CharmRepo.build_from_config')
    @mock.patch('mojo.charm_repo.CharmRepo.build_all')
    def test_auto_repo(self, build_all, build_from_config):
        phase = CharmRepoPhase(auto_repo=True)
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['repo', 'services.cfg'], stage='staging')

        phase.run('unused', workspace, 'trusty', 'staging')
        build_all.assert_called_with(
            os.path.join(os.path.dirname(workspace.spec_dir), 'build'),
            'trusty')
